# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""PKI CLI."""

import logging
import os
import sys
from pathlib import Path

from click import Path as ClickPath
from click import echo, group, option, pass_context, password_option

from openfl.utilities.ca.ca import (
    CA_CONFIG_JSON,
    CA_PASSWORD_FILE,
    CA_PKI_DIR,
    CA_STEP_CONFIG_DIR,
    certify,
    get_ca_bin_paths,
    get_token,
    install,
    remove_ca,
    run_ca,
)

logger = logging.getLogger(__name__)

CA_URL = "localhost:9123"


@group()
@pass_context
def pki(context):
    """Manage Step-CA PKI."""
    context.obj["group"] = "pki"


@pki.command(name="run")
@option("-p", "--ca-path", required=True, help="The ca path", type=ClickPath())
def run_(ca_path):
    """Starts a CA server."""
    run(ca_path)


def run(ca_path):
    """Starts a CA server.

    Args:
        ca_path (str): The ca path.
    """
    ca_path = Path(ca_path).absolute()
    step_config_dir = ca_path / CA_STEP_CONFIG_DIR
    pki_dir = ca_path / CA_PKI_DIR
    password_file = pki_dir / CA_PASSWORD_FILE
    ca_json = step_config_dir / CA_CONFIG_JSON
    _, step_ca_path = get_ca_bin_paths(ca_path)
    if (
        not os.path.exists(step_config_dir)
        or not os.path.exists(pki_dir)
        or not os.path.exists(password_file)
        or not os.path.exists(ca_json)
        or not os.path.exists(step_ca_path)
    ):
        logger.error("CA is not installed or corrupted, please install it first")
        sys.exit(1)
    run_ca(step_ca_path, password_file, ca_json)


@pki.command(name="install")
@option("-p", "--ca-path", required=True, help="Path to CA.", type=ClickPath())
@password_option(prompt="The password will encrypt CA files. \nEnter the password: ")
@option("--ca-url", required=False, default=CA_URL, show_default=True)
def install_(ca_path, password, ca_url):
    """Creates a CA workspace, optionally password protected."""
    ca_path = Path(ca_path).absolute()
    install(ca_path, ca_url, password)


@pki.command(name="uninstall")
@option("-p", "--ca-path", required=True, help="Path to CA to be uninstalled.", type=ClickPath())
def uninstall(ca_path):
    """Removes Step-CA."""
    ca_path = Path(ca_path).absolute()
    remove_ca(ca_path)


@pki.command(name="get-token")
@option("-n", "--name", required=True)
@option(
    "--ca-url", required=False, default=CA_URL, help="Full URL of CA server.", show_default=True
)
@option(
    "-p",
    "--ca-path",
    default=".",
    help="Path to CA binaries, defaults to current directory.",
    type=ClickPath(exists=True),
)
def get_token_(name, ca_url, ca_path):
    """Creates an authentication token."""
    ca_path = Path(ca_path).absolute()
    token = get_token(name, ca_url, ca_path)
    echo(f"Token: {token}")


@pki.command(name="certify")
@option(
    "-n",
    "--name",
    required=True,
    help=(
        "Subject Alternative Name (SAN) to use for certificate. "
        "Use FQDN for aggregator, and common name for collaborator"
    ),
)
@option("-t", "--token", "token_with_cert", required=True, help="Authentication token.")
@option(
    "-c",
    "--certs-path",
    required=False,
    default=Path(".") / "cert",
    help="The path where certificates will be stored",
    type=ClickPath(),
)
@option(
    "-p",
    "--ca-path",
    default=".",
    help="Path to CA client, defaults to current directory.",
    type=ClickPath(exists=True),
    required=False,
    show_default=True,
)
def certify_(name, token_with_cert, certs_path, ca_path):
    """Generates a certificate for the given name."""
    certs_path = Path(certs_path).absolute()
    ca_path = Path(ca_path).absolute()
    certs_path.mkdir(parents=True, exist_ok=True)
    certify(name, certs_path, token_with_cert, ca_path)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


import functools
from typing import Callable

from openfl.experimental.workflow.utilities import RedirectStdStreamContext


def aggregator(f: Callable = None) -> Callable:
    """Placement decorator that designates that the task will run at the
    aggregator node.

    Usage:
    class MyFlow(FLSpec):
        ...
        @aggregator
        def agg_task(self):
           ...
        ...

    Args:
        f (Callable, optional): The function to be decorated.

    Returns:
        Callable: The decorated function.
    """
    print(f'Aggregator step "{f.__name__}" registered')
    f.is_step = True
    f.decorators = []
    f.name = f.__name__
    f.task = True
    f.aggregator_step = True
    f.collaborator_step = False
    if f.__doc__:
        f.__doc__ = "<Node: Aggregator>" + f.__doc__

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        print(f"\nCalling {f.__name__}")
        with RedirectStdStreamContext() as context_stream:
            # context_stream capture stdout and stderr for the function f.__name__
            wrapper._stream_buffer = context_stream
            f(*args, **kwargs)

    return wrapper


def collaborator(f: Callable = None) -> Callable:
    """Placement decorator that designates that the task will run at the
    collaborator node.

    Usage:
    class MyFlow(FLSpec):
        ...
        @collaborator
        def collaborator_task(self):
            ...

        @collaborator(num_gpus=1)
        @def collaborator_gpu_task:
            ...
        ...

    Args:
        f (Callable, optional): The function to be decorated.
        num_gpus (float, optional): [Applicable for Ray backend only]
                  Defines how many GPUs will be made available
                  to the task (Default = 0). Selecting a value < 1 (0.0-1.0]
                  will result in sharing of GPUs between tasks. 1 >= results in
                  exclusive GPU access for the task.

    Returns:
        Callable: The decorated function.
    """
    if f is None:
        return functools.partial(collaborator)

    print(f'Collaborator step "{f.__name__}" registered')
    f.is_step = True
    f.decorators = []
    f.name = f.__name__
    f.task = True
    f.aggregator_step = False
    f.collaborator_step = True
    if f.__doc__:
        f.__doc__ = "<Node: Collaborator>" + f.__doc__

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        print(f"\nCalling {f.__name__}")
        with RedirectStdStreamContext() as context_stream:
            # context_stream capture stdout and stderr for the function f.__name__
            wrapper._stream_buffer = context_stream
            f(*args, **kwargs)

    return wrapper



# Copyright (C) 2024 Intel Corporation
# Licensed subject to the terms of the separately executed evaluation license agreement between
# Intel Corporation and you.

from openfl.federated import XGBoostDataLoader
import os
import modin.pandas as pd

class HiggsDataLoader(XGBoostDataLoader):
    """
    DataLoader for the Higgs dataset.

    This class inherits from XGBoostDataLoader and is responsible for loading
    the Higgs dataset for training and validation.

    Attributes:
        X_train (numpy.ndarray): Training features.
        y_train (numpy.ndarray): Training labels.
        X_valid (numpy.ndarray): Validation features.
        y_valid (numpy.ndarray): Validation labels.
    """
    def __init__(self, data_path, **kwargs):
        super().__init__(**kwargs)
        X_train, y_train, X_valid, y_valid = load_Higgs(
            data_path, **kwargs
        )
        self.X_train = X_train
        self.y_train = y_train
        self.X_valid = X_valid
        self.y_valid = y_valid


def load_Higgs(data_path, **kwargs):
    """
    Load the Higgs dataset from CSV files.

    The dataset is expected to be in two CSV files: 'train.csv' and 'valid.csv'.
    The first column in each file represents the labels, and the remaining
    columns represent the features.

    Args:
        data_path (str): The directory path where the CSV files are located.
        **kwargs: Additional keyword arguments.

    Returns:
        tuple: A tuple containing four elements:
            - X_train (numpy.ndarray): Training features.
            - y_train (numpy.ndarray): Training labels.
            - X_valid (numpy.ndarray): Validation features.
            - y_valid (numpy.ndarray): Validation labels.
    """
    train_data = pd.read_csv(os.path.join(data_path, 'train.csv'), header=None)
    X_train = train_data.iloc[:, 1:].values
    y_train = train_data.iloc[:, 0].values

    valid_data = pd.read_csv(os.path.join(data_path, 'valid.csv'), header=None)
    X_valid = valid_data.iloc[:, 1:].values
    y_valid = valid_data.iloc[:, 0].values

    return X_train, y_train, X_valid, y_valid



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.interface package."""

from openfl.experimental.workflow.interface.fl_spec import FLSpec
from openfl.experimental.workflow.interface.participants import Aggregator, Collaborator



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Data package."""

from importlib import util

from openfl.federated.data.loader import DataLoader  # NOQA

if util.find_spec("keras") is not None:
    from openfl.federated.data.loader_keras import KerasDataLoader  # NOQA

if util.find_spec("torch") is not None:
    from openfl.federated.data.loader_pt import PyTorchDataLoader  # NOQA

if util.find_spec("xgboost") is not None:
    from openfl.federated.data.loader_xgb import XGBoostDataLoader  # NOQA



import re

def is_safe_path(path):
    """
    Validate the path to ensure it contains only allowed characters.

    Args:
        path (str): The path to validate.

    Returns:
        bool: True if the path is safe, False otherwise.
    """
    return re.match(r'^[\w\-/\.]+$', path) is not None



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import os
import time

from tensorboardX import SummaryWriter

from openfl.callbacks.callback import Callback

logger = logging.getLogger(__name__)


class MetricWriter(Callback):
    """Log scalar metrics at the end of each round.

    Attributes:
        log_dir: Path to write logs as lines of JSON. Defaults to `./logs`.
        use_tensorboard: If True, writes scalar summaries to TensorBoard under `log_dir`.
    """

    def __init__(self, log_dir: str = "./logs/", use_tensorboard: bool = True):
        super().__init__()
        self.log_dir = log_dir
        self.use_tensorboard = use_tensorboard

        self._log_file_handle = None
        self._summary_writer = None
        self._round_start_time = None

    def on_experiment_begin(self, logs=None):
        """Open file handles for logging."""
        os.makedirs(self.log_dir, exist_ok=True)

        if not self._log_file_handle:
            self._log_file_handle = open(
                os.path.join(self.log_dir, self.params["origin"] + "_metrics.txt"), "a"
            )

        if self.use_tensorboard:
            self._summary_writer = SummaryWriter(
                os.path.join(self.log_dir, self.params["origin"] + "_tensorboard")
            )

    def on_round_end(self, round_num: int, logs=None):
        """Log metrics.

        Args:
            round_num: The current round number.
            logs: A key-value pair of scalar metrics.
        """
        logs = logs or {}
        elapsed_seconds = time.monotonic() - self._round_start_time
        metrics = {
            "round_number": round_num,
            "elapsed_seconds": elapsed_seconds,
            **logs,
        }
        logger.info(f"Round {round_num}: Metrics: {metrics}")

        self._log_file_handle.write(json.dumps(metrics) + "\n")
        self._log_file_handle.flush()

        if self._summary_writer:
            for key, value in metrics.items():
                self._summary_writer.add_scalar(key, value, round_num)
            self._summary_writer.flush()

    def on_experiment_end(self, logs=None):
        """Cleanup."""
        if self._log_file_handle:
            self._log_file_handle.close()
            self._log_file_handle = None

        if self._summary_writer:
            self._summary_writer.close()
            
    def on_round_begin(self, round_num: int, logs=None):
        self._round_start_time = time.monotonic()


# Copyright 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""
This file contains callback that help setup for secure aggregation for the
collaborator.
"""

import json
import logging
import struct

import numpy as np

from openfl.callbacks.callback import Callback
from openfl.protocols import utils
from openfl.utilities import TensorKey

logger = logging.getLogger(__name__)


class SecAggBootstrapping(Callback):
    """
    This callback is used by the collaborator to perform secure aggregation
    bootstrapping.

    Required params include:
    - origin: Name of the collaborator using the callback.
    - client: AggregatorGRPCClient to communicate with the aggregator server.

    It also requires the tensor-db client to be set.
    """

    def __init__(self):
        super().__init__()

        from importlib import util

        # Check if pycryptodome is installed.
        if util.find_spec("Crypto") is None:
            raise Exception(
                "'pycryptodome' not installed."
                "This package is necessary when secure aggregation is enabled."
            )

    def on_experiment_begin(self, logs=None):
        """
        Used to perform secure aggregation setup before experiment begins.
        """
        self.name = self.params["origin"]
        self.client = self.params["client"]
        logger.info("Secure aggregation is enabled, starting setup...")
        # Generate private public key pair used for secure aggregation.
        self._generate_keys()
        # Fetch public keys for all collaborators from the aggregator.
        collaborator_keys = self._fetch_public_keys()
        # Generate ciphertexts for each collaborator and share them with the
        # aggregator.
        self._generate_ciphertexts(collaborator_keys)
        # Decrypt the addressed ciphertexts and share them with the
        # aggregator.
        self._decrypt_ciphertexts(collaborator_keys)
        # Save the tensors which are required for masking of gradients.
        self._save_mask_tensors()

    def _generate_keys(self):
        """
        Generates a pair of private and public keys, along with a private seed,
        and updates the local and global results.

        This method performs the following steps:
        1. Generates two pairs of private and public keys.
        2. Creates a local result dictionary containing the private keys,
            public keys, and a private seed.
        3. Creates a global result dictionary containing the public keys.
        4. Sends the global results to the participant via the aggregator's
            secure aggregation mechanism.
        5. Updates the instance parameters with the local result.
        """
        from openfl.utilities.secagg import generate_key_pair

        private_key1, public_key1 = generate_key_pair()
        private_key2, public_key2 = generate_key_pair()

        local_result = {
            "private_key": [private_key1, private_key2],
            "public_key": [public_key1, public_key2],
            "private_seed": np.random.random(),
        }
        global_results = {
            "public_key": [public_key1, public_key2],
        }

        self._send_to_aggregator(global_results, "generate_keys")
        # Update callback params as the results for this step are reused at a
        # later stage.
        self.params.update(local_result)
        logger.debug("SecAgg: Generate key-pair generation successful")

    def _fetch_public_keys(self):
        """
        Fetches collaborators' public keys from the aggregator and identifies
        the index of the current collaborator using it's public key.

        Returns:
            dict: A dictionary containing the public keys of all collaborators,
                where the keys are the collaborator indices and the values are
                the public keys.
        """
        public_keys = {}
        public_keys_tensor = self._fetch_from_aggregator("public_keys")
        for tensor in public_keys_tensor:
            # Creating a dictionary of the received public keys.
            public_keys[int(tensor[0])] = [tensor[1], tensor[2]]
            # Finding the index of the current collaborator by matching the
            # first public key.
            if tensor[1] == self.params["public_key"][0]:
                self.index = int(tensor[0])

        return public_keys

    def _generate_ciphertexts(self, public_keys):
        """
        Generate ciphertexts for secure aggregation.

        This method generates ciphertexts for each collaborator using their
        public keys. It creates secret shares for the private seed and private
        key, then uses these shares to generate agreed keys and ciphertexts
        for secure communication between collaborators.

        Args:
            public_keys (dict): A dictionary where keys are collaborator
                indices and values are lists containing public keys of the
                collaborators.
        """
        from openfl.utilities.secagg import (
            create_ciphertext,
            create_secret_shares,
            generate_agreed_key,
        )

        logger.debug("SecAgg: Generating ciphertexts to be shared with other collaborators")
        collaborator_count = len(public_keys)

        private_seed = self.params["private_seed"]
        seed_shares = create_secret_shares(
            # Converts the floating-point number private_seed into an 8-byte
            # binary representation.
            struct.pack("d", private_seed),
            collaborator_count,
            collaborator_count,
        )

        private_keys = self.params["private_key"]
        # Create secret shares for the private key.
        key_shares = create_secret_shares(
            str.encode(private_keys[0]),
            collaborator_count,
            collaborator_count,
        )

        global_results = {"ciphertext": []}
        local_result = {"ciphertext_verification": {}, "agreed_keys": []}
        # Create cipher-texts for each collaborator.
        for collab_index in public_keys:
            agreed_key = generate_agreed_key(private_keys[0], public_keys[collab_index][0])
            ciphertext, mac, nonce = create_ciphertext(
                agreed_key,  # agreed key
                self.index,  # source collaborator index
                collab_index,  # destination collaborator index
                seed_shares[collab_index],  # seed share from source to dest
                key_shares[collab_index],  # key share from source to dest
            )
            global_results["ciphertext"].append((self.index, collab_index, str(ciphertext)))
            local_result["ciphertext_verification"][collab_index] = [ciphertext, mac, nonce]
            local_result["agreed_keys"].append([self.index, collab_index, agreed_key])

        self._send_to_aggregator(global_results, "generate_ciphertexts")
        # Update callback params as the results for this step are reused at a
        # later stage.
        self.params.update(local_result)

        logger.debug("SecAgg: Ciphertexts shared with the aggregator successfully")

    def _decrypt_ciphertexts(self, public_keys):
        """
        Decrypts the ciphertexts received from collaborators using the provided
        public keys.

        This method fetches the ciphertexts from the aggregator, decrypts them
        using the collaborator's private key and the provided public keys, and
        then sends the decrypted seed shares and key shares back to the
        aggregator.

        Args:
            public_keys (dict): A dictionary containing the public keys of the
                collaborators.
        """
        from openfl.utilities.secagg import (
            decipher_ciphertext,
            generate_agreed_key,
        )

        logger.debug("SecAgg: fetching addressed ciphertexts from the aggregator")

        ciphertexts = self._fetch_from_aggregator("ciphertexts")
        private_keys = self.params["private_key"]
        ciphertext_verification = self.params["ciphertext_verification"]

        global_results = {"seed_share": [], "key_share": []}

        for cipher in ciphertexts:
            source_index = int(cipher[0])
            if int(cipher[1]) == self.index:
                _, _, seed_share, key_share = decipher_ciphertext(
                    generate_agreed_key(private_keys[0], public_keys[source_index][0]),
                    ciphertext_verification[source_index][0],
                    ciphertext_verification[source_index][1],
                    ciphertext_verification[source_index][2],
                )
                global_results["seed_share"].append((source_index, self.index, str(seed_share)))
                global_results["key_share"].append((source_index, self.index, str(key_share)))

        self._send_to_aggregator(global_results, "decrypt_ciphertexts")

        logger.debug("SecAgg: decrypted ciphertexts shared with the aggregator")

    def _generate_masks(self):
        """
        Use the private seed and agreed keys to calculate the masks to be
        added to the gradients.
        """
        from openfl.utilities.secagg import (
            calculate_shared_mask,
            pseudo_random_generator,
        )

        private_mask = pseudo_random_generator(self.params.get("private_seed"))
        shared_mask = calculate_shared_mask(self.params.get("agreed_keys"))

        return private_mask, shared_mask

    def _save_mask_tensors(self):
        """
        Generates private and shared masks, stores them in a local tensor
        dictionary, and caches the dictionary in the tensor database.

        These tensors are then added to the gradient before sharing them
        with the aggregator during training task.
        """
        private_mask, shared_mask = self._generate_masks()
        local_tensor_dict = {
            TensorKey("private_mask", self.name, -1, False, ("secagg",)): [private_mask],
            TensorKey("shared_mask", self.name, -1, False, ("secagg",)): [shared_mask],
        }
        self.tensor_db.cache_tensor(local_tensor_dict)
        logger.info("SecAgg: setup completed, saved required tensors to db.")

    def _send_to_aggregator(self, tensor_dict: dict, stage: str):
        """
        Sends the provided tensor dictionary to the aggregator after
        compressing it.

        Args:
            tensor_dict (dict): A dictionary where keys are tensor names and
                values are numpy arrays.
            stage (str): The current stage of the secure aggregation process.
        """
        named_tensors = []
        # Convert python dict to tensor dict.
        for key, nparray in tensor_dict.items():
            tensor_key = TensorKey(
                key,
                self.name,
                -1,
                False,
                (
                    self.name,
                    "secagg",
                ),
            )
            named_tensor = utils.construct_named_tensor(
                tensor_key, str.encode(json.dumps(nparray)), {}, lossless=True
            )
            named_tensors.append(named_tensor)

        self.client.send_local_task_results(-1, f"secagg_{stage}", -1, named_tensors)

    def _fetch_from_aggregator(self, key_name):
        """
        Fetches the aggregated tensor data from a aggregator.

        Args:
            key_name (str): The name of the key to fetch the tensor for.

        Returns:
            bytes: The aggregated tensor data in bytes.
        """
        tensor = self.client.get_aggregated_tensor(key_name, -1, False, ("secagg",), True)
        return json.loads(tensor.data_bytes)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Aggregator module."""

import sys
from logging import getLogger
from pathlib import Path

from click import (
    Path as ClickPath,
)
from click import (
    confirm,
    echo,
    group,
    option,
    pass_context,
    style,
)

from openfl.cryptography.ca import sign_certificate
from openfl.cryptography.io import (
    get_csr_hash,
    read_crt,
    read_csr,
    read_key,
    write_crt,
    write_key,
)
from openfl.cryptography.participant import generate_csr
from openfl.federated import Plan
from openfl.interface.cli_helper import CERT_DIR
from openfl.utilities import click_types
from openfl.utilities.path_check import is_directory_traversal
from openfl.utilities.utils import getfqdn_env

logger = getLogger(__name__)


@group()
@pass_context
def aggregator(context):
    """Manage Federated Learning Aggregator."""
    context.obj["group"] = "aggregator"


@aggregator.command(name="start")
@option(
    "-p",
    "--plan",
    required=False,
    help="Path to an FL plan.",
    default="plan/plan.yaml",
    type=ClickPath(exists=True),
    show_default=True,
)
@option(
    "-c",
    "--authorized_cols",
    required=False,
    help="Path to an authorized collaborator list.",
    default="plan/cols.yaml",
    type=ClickPath(exists=True),
    show_default=True,
)
@option(
    "--task_group",
    required=False,
    help="Task group to execute as defined in the plan task assigner.",
)
def start_(plan, authorized_cols, task_group):
    """Starts the aggregator service."""
    if is_directory_traversal(plan):
        echo("Federated learning plan path is out of the openfl workspace scope.")
        sys.exit(1)
    if is_directory_traversal(authorized_cols):
        echo("Authorized collaborator list file path is out of the openfl workspace scope.")
        sys.exit(1)

    # Parse plan and override mode if specified
    parsed_plan = Plan.parse(
        plan_config_path=Path(plan).absolute(),
        cols_config_path=Path(authorized_cols).absolute(),
    )

    # Set task_group in assigner settings if provided
    if task_group:
        if "settings" not in parsed_plan.config["assigner"]:
            parsed_plan.config["assigner"]["settings"] = {}
        parsed_plan.config["assigner"]["settings"]["selected_task_group"] = task_group
        logger.info(f"Setting aggregator to assign: {task_group} task_group")

    logger.info("🧿 Starting the Aggregator Service.")

    parsed_plan.get_server().serve()


@aggregator.command(name="generate-cert-request")
@option(
    "--fqdn",
    required=False,
    type=click_types.FQDN,
    help="The fully qualified domain name of aggregator node.",
)
def _generate_cert_request(fqdn):
    """Generates aggregator certificate key-pair."""
    generate_cert_request(fqdn)


def generate_cert_request(fqdn):
    if fqdn is None:
        fqdn = getfqdn_env()

    common_name = f"{fqdn}".lower()
    subject_alternative_name = f"DNS:{common_name}"
    file_name = f"agg_{common_name}"

    echo(
        f"Creating AGGREGATOR certificate key pair with following settings: "
        f"CN={style(common_name, fg='red')},"
        f" SAN={style(subject_alternative_name, fg='red')}"
    )

    server_private_key, server_csr = generate_csr(common_name, server=True)

    (CERT_DIR / "server").mkdir(parents=True, exist_ok=True)

    echo("  Writing AGGREGATOR certificate key pair to: " + style(f"{CERT_DIR}/server", fg="green"))

    # Print csr hash before writing csr to disk
    csr_hash = get_csr_hash(server_csr)
    echo("The CSR Hash " + style(f"{csr_hash}", fg="red"))

    # Write aggregator csr and key to disk
    write_crt(server_csr, CERT_DIR / "server" / f"{file_name}.csr")
    write_key(server_private_key, CERT_DIR / "server" / f"{file_name}.key")


# TODO: function not used
def find_certificate_name(file_name):
    """Search the CRT for the actual aggregator name.

    Args:
        file_name (str): The name of the file to search.

    Returns:
        str: The name of the aggregator found in the CRT.
    """
    # This loop looks for the collaborator name in the key
    with open(file_name, "r", encoding="utf-8") as f:
        for line in f:
            if "Subject: CN=" in line:
                col_name = line.split("=")[-1].strip()
                break
    return col_name


@aggregator.command(name="certify")
@option(
    "-n",
    "--fqdn",
    type=click_types.FQDN,
    help="The fully qualified domain name of aggregator node.",
)
@option("-s", "--silent", help="If set, skips manual confirmation.", is_flag=True)
def _certify(fqdn, silent):
    """Certifies the aggregator certificate key pair."""
    certify(fqdn, silent)


def certify(fqdn, silent):
    """Certifies the aggregator certificate key pair.

    Args:
        fqdn (str): The fully qualified domain name of aggregator node.
        silent (bool): Flag to enable silent mode.
    """

    if fqdn is None:
        fqdn = getfqdn_env()

    common_name = f"{fqdn}".lower()
    file_name = f"agg_{common_name}"
    cert_name = f"server/{file_name}"
    signing_key_path = "ca/signing-ca/private/signing-ca.key"
    signing_crt_path = "ca/signing-ca.crt"

    # Load CSR
    csr_path_absolute_path = Path(CERT_DIR / f"{cert_name}.csr").absolute()
    if not csr_path_absolute_path.exists():
        echo(
            style("Aggregator certificate signing request not found.", fg="red")
            + " Please run `fx aggregator generate-cert-request`"
            " to generate the certificate request."
        )

    csr, csr_hash = read_csr(csr_path_absolute_path)

    # Load private signing key
    private_sign_key_absolute_path = Path(CERT_DIR / signing_key_path).absolute()
    if not private_sign_key_absolute_path.exists():
        echo(
            style("Signing key not found.", fg="red") + " Please run `fx workspace certify`"
            " to initialize the local certificate authority."
        )

    signing_key = read_key(private_sign_key_absolute_path)

    # Load signing cert
    signing_crt_absolute_path = Path(CERT_DIR / signing_crt_path).absolute()
    if not signing_crt_absolute_path.exists():
        echo(
            style("Signing certificate not found.", fg="red") + " Please run `fx workspace certify`"
            " to initialize the local certificate authority."
        )

    signing_crt = read_crt(signing_crt_absolute_path)

    echo(
        "The CSR Hash for file "
        + style(f"{cert_name}.csr", fg="green")
        + " = "
        + style(f"{csr_hash}", fg="red")
    )

    crt_path_absolute_path = Path(CERT_DIR / f"{cert_name}.crt").absolute()

    if silent:
        echo(" Warning: manual check of certificate hashes is bypassed in silent mode.")
        echo(" Signing AGGREGATOR certificate")
        signed_agg_cert = sign_certificate(csr, signing_key, signing_crt.subject)
        write_crt(signed_agg_cert, crt_path_absolute_path)

    else:
        echo("Make sure the two hashes above are the same.")
        if confirm("Do you want to sign this certificate?"):
            echo(" Signing AGGREGATOR certificate")
            signed_agg_cert = sign_certificate(csr, signing_key, signing_crt.subject)
            write_crt(signed_agg_cert, crt_path_absolute_path)

        else:
            echo(
                style("Not signing certificate.", fg="red")
                + " Please check with this AGGREGATOR to get the correct"
                " certificate for this federation."
            )



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Experimental CLI."""

import os
from pathlib import Path

from click import group, pass_context


@group()
@pass_context
def experimental(context):
    """Manage Experimental Environment."""
    context.obj["group"] = "experimental"


@experimental.command(name="deactivate")
def deactivate():
    """Deactivate experimental environment."""
    settings = Path("~").expanduser().joinpath(".openfl", "experimental").resolve()

    os.remove(settings)



# Copyright (C) 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from keras.models import Sequential
from keras.layers import Conv2D
from keras.layers import Dense
from keras.layers import Flatten

from openfl.federated import KerasTaskRunner

import logging
logger = logging.getLogger(__name__)

class KerasCNN(KerasTaskRunner):
    """A basic convolutional neural network model."""

    def __init__(self, **kwargs):
        """
        Initialize.

        Args:
            **kwargs: Additional parameters to pass to the function
        """
        super().__init__(**kwargs)

        self.model = self.build_model(self.feature_shape, self.data_loader.num_classes, **kwargs)

        self.initialize_tensorkeys_for_functions()

        self.model.summary(print_fn=logger.info)

        logger.info(f'Train Set Size : {self.get_train_data_size()}')
        logger.info(f'Valid Set Size : {self.get_valid_data_size()}')

    def build_model(self,
                    input_shape,
                    num_classes,
                    conv_kernel_size=(4, 4),
                    conv_strides=(2, 2),
                    conv1_channels_out=16,
                    conv2_channels_out=32,
                    final_dense_inputsize=100,
                    **kwargs):
        """
        Define the model architecture.

        Args:
            input_shape (numpy.ndarray): The shape of the data
            num_classes (int): The number of classes of the dataset

        Returns:
            keras.models.Sequential: The model defined in Keras

        """
        model = Sequential()

        model.add(Conv2D(conv1_channels_out,
                         kernel_size=conv_kernel_size,
                         strides=conv_strides,
                         activation='relu',
                         input_shape=input_shape))

        model.add(Conv2D(conv2_channels_out,
                         kernel_size=conv_kernel_size,
                         strides=conv_strides,
                         activation='relu'))

        model.add(Flatten())

        model.add(Dense(final_dense_inputsize, activation='relu'))

        model.add(Dense(num_classes, activation='softmax'))

        model.compile(loss="categorical_crossentropy",
                      optimizer="adam",
                      metrics=["accuracy"])

        return model



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
from openfl.callbacks.callback import Callback


class LambdaCallback(Callback):
    """Custom on-the-fly callbacks.

    This callback can be constructed with functions that will be called
    at the appropriate time during the life-cycle of a Federated Learning experiment.
    Certain callbacks may expect positional arguments, for example:

    * on_round_begin: expects `round_num` as a positional argument.
    * on_round_end: expects `round_num` as a positional argument.

    Args:
        on_round_begin: called at the beginning of every round.
        on_round_end: called at the end of every round.
        on_experiment_begin: called at the beginning of an experiment.
        on_experiment_end: called at the end of an experiment.
    """

    def __init__(
        self,
        on_round_begin=None,
        on_round_end=None,
        on_experiment_begin=None,
        on_experiment_end=None,
    ):
        super().__init__()
        if on_round_begin is not None:
            self.on_round_begin = on_round_begin
        if on_round_end is not None:
            self.on_round_end = on_round_end
        if on_experiment_begin is not None:
            self.on_experiment_begin = on_experiment_begin
        if on_experiment_end is not None:
            self.on_experiment_end = on_experiment_end



import grpc
from flwr.proto import grpcadapter_pb2_grpc
from src.grpc.connector.flower.message_conversion import flower_to_openfl_message, openfl_to_flower_message
from logging import getLogger

class FlowerInteropClient:
    """
    FlowerInteropClient facilitates communication between the Flower SuperLink
    and the OpenFL Server. It converts messages between OpenFL and Flower formats
    and handles the send-receive communication with the Flower SuperNode using gRPC.
    """
    def __init__(self, superlink_address, automatic_shutdown=False):
        """
        Initialize.

        Args:
            superlink_address: The address the Flower SuperLink will listen on
        """
        self.superlink_channel = grpc.insecure_channel(superlink_address)
        self.superlink_stub = grpcadapter_pb2_grpc.GrpcAdapterStub(self.superlink_channel)

        self.automatic_shutdown = automatic_shutdown
        self.end_experiment = False
        self.is_flwr_serverapp_running_callback = None

        self.logger = getLogger(__name__)

    def set_is_flwr_serverapp_running_callback(self, is_flwr_serverapp_running_callback):
        self.is_flwr_serverapp_running_callback = is_flwr_serverapp_running_callback

    def send_receive(self, openfl_message, header):
        """
        Sends a message to the Flower SuperLink and receives the response.

        Args:
            openfl_message: converted Flower SuperNode request sent by OpenFL server
            header: OpenFL header information to be included in the message.

        Returns:
            The response from the Flower SuperLink, converted back to OpenFL format.
        """
        flower_message = openfl_to_flower_message(openfl_message)
        flower_response = self.superlink_stub.SendReceive(flower_message)

        if self.automatic_shutdown and self.is_flwr_serverapp_running_callback:
            # Check if the flwr_serverapp subprocess is still running, if it isn't
            # then the experiment has completed
            self.end_experiment = not self.is_flwr_serverapp_running_callback()

        openfl_response = flower_to_openfl_message(flower_response,
                                                   header=header,
                                                   end_experiment=self.end_experiment)

        return openfl_response



# Copyright (C) 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from openfl.federated import KerasDataLoader
import numpy as np
from sklearn.model_selection import train_test_split
from glob import glob


class KerasHippmapp3rsynth(KerasDataLoader):
    """Data Loader for synthetic Hippmapp3r Dataset."""

    def __init__(self, data_path, batch_size, **kwargs):
        """
        Initialize.

        Args:
            data_path: File path for the dataset
            batch_size (int): The batch size for the data loader
            **kwargs: Additional arguments, passed to super init and load_mnist_shard
        """
        super().__init__(batch_size, **kwargs)

        X_train = glob(f"{data_path}/X*.npy")
        y_train = glob(f"{data_path}/y*.npy")
        # Perform test-train split
        X_train, X_valid, y_train, y_valid = train_test_split(
            X_train, y_train, test_size=0.2, random_state=42
        )
        self.X_train = np.asarray(X_train)
        self.X_valid = np.asarray(X_valid)
        self.y_train = np.asarray(y_train)
        self.y_valid = np.asarray(y_valid)

    @staticmethod
    def _batch_generator(X, y, idxs, batch_size, num_batches):
        """Generates batches of data.

        Args:
            X (np.array): The input data.
            y (np.array): The label data.
            idxs (np.array): The index of the dataset.
            batch_size (int): The batch size for the data loader.
            num_batches (int): The number of batches.

        Yields:
            tuple: The input data and label data for each batch.
        """
        for i in range(num_batches):
            a = i * batch_size
            b = a + batch_size
            x_list = []
            y_list = []
            for _x, _y in zip(X[idxs[a:b]], y[idxs[a:b]]):
                x_list.append(np.load(_x))
                y_list.append(np.load(_y))
            yield np.stack(x_list), np.stack(y_list)



from logging import getLogger
logger = getLogger(__name__)

import psutil
import subprocess
import sys
import signal

from src.grpc.connector.flower.interop_client import FlowerInteropClient
from src.util import is_safe_path

import os

flwr_home = os.path.join(os.getcwd(), "save/.flwr")
if not is_safe_path(flwr_home):
    raise ValueError("Invalid path for FLWR_HOME")

os.environ["FLWR_HOME"] = flwr_home
os.makedirs(os.environ["FLWR_HOME"], exist_ok=True)

class ConnectorFlower:
    """
    A Connector subclass specifically designed for integrating with the Flower framework.
    This class is responsible for constructing and managing the execution of Flower server commands.
    """

    def __init__(self,
                 superlink_params: dict,
                 flwr_run_params: dict = None,
                 automatic_shutdown: bool = True,
                 **kwargs):
        """
        Initialize the ConnectorFlower instance by setting up the necessary server commands.

        Args:
            superlink_params (dict): Configuration settings for the Flower server.
            flwr_run_params (dict, optional): Parameters for running the Flower application.
            automatic_shutdown (bool, optional): Flag to enable automatic shutdown of the server. Defaults to True.
            **kwargs: Additional keyword arguments.
        """
        super().__init__()
        self._process = None

        self.automatic_shutdown = automatic_shutdown
        self.signal_shutdown_sent = False

        self.superlink_params = superlink_params
        self.flwr_superlink_command = self._build_flwr_superlink_command()

        self.flwr_run_params = flwr_run_params
        self.flwr_run_command = self._build_flwr_run_command() if self.flwr_run_params else None

        self.interop_client = None
        signal.signal(signal.SIGINT, self._handle_sigint)

    def get_interop_client(self):
        """
        Create and return a LocalGRPCClient instance using the superlink parameters.

        Returns:
            LocalGRPCClient: An instance configured with the connector address and server rounds.
        """
        connector_address = self.superlink_params.get("fleet-api-address", "0.0.0.0:9092")
        self.interop_client = FlowerInteropClient(connector_address, self.automatic_shutdown)
        return self.interop_client

    def _build_flwr_superlink_command(self) -> list[str]:
        """
        Construct the command to initiate the Flower SuperLink based on provided parameters.

        Returns:
            list[str]: A list of command-line arguments for starting the Flower server.
        """

        command = ["flower-superlink", "--fleet-api-type", "grpc-adapter"]

        if "insecure" in self.superlink_params and self.superlink_params["insecure"]:
            command += ["--insecure"]

        if "serverappio-api-address" in self.superlink_params:
            command += ["--serverappio-api-address", str(self.superlink_params["serverappio-api-address"])]
            # flwr default: 0.0.0.0:9091

        if "fleet-api-address" in self.superlink_params:
            command += ["--fleet-api-address", str(self.superlink_params["fleet-api-address"])]
            # flwr default: 0.0.0.0:9092

        if "exec-api-address" in self.superlink_params:
            command += ["--exec-api-address", str(self.superlink_params["exec-api-address"])]
            # flwr default: 0.0.0.0:9093

        if self.automatic_shutdown:
            command += ["--isolation", "process"]
            self.flwr_serverapp_command = self._build_flwr_serverapp_command()
            # flwr will default to "--isolation subprocess"

        return command

    def _build_flwr_serverapp_command(self) -> list[str]:
        """
        Construct the command to start the Flower ServerApp based on superlink parameters.

        Returns:
            list[str]: A list of command-line arguments for starting the Flower ServerApp.
        """
        command = ["flwr-serverapp", "--run-once"]

        if "insecure" in self.superlink_params and self.superlink_params["insecure"]:
            command += ["--insecure"]

        if "serverappio-api-address" in self.superlink_params:
            command += ["--serverappio-api-address", str(self.superlink_params["serverappio-api-address"])]

        return command

    def is_flwr_serverapp_running(self):
        """
        Determine if the Flower ServerApp subprocess is currently active.

        Returns:
            bool: True if the ServerApp is running, False otherwise.
        """
        if not hasattr(self, 'flwr_serverapp_subprocess'):
            logger.debug("[OpenFL Connector] ServerApp was never started.")
            return False

        if self.flwr_serverapp_subprocess.poll() is None:
            logger.debug("[OpenFL Connector] ServerApp is still running.")
            return True

        if not self.signal_shutdown_sent:
            self.signal_shutdown_sent = True
            logger.info("[OpenFL Connector] Experiment has ended. Sending signal to shut down Flower components.")

        return False

    def _stop_flwr_serverapp(self):
        """Terminate the `flwr_serverapp` subprocess if it is still active."""
        if hasattr(self, 'flwr_serverapp_subprocess') and self.flwr_serverapp_subprocess.poll() is None:
            logger.debug("[OpenFL Connector] ServerApp still running. Stopping...")
            self.flwr_serverapp_subprocess.terminate()
            try:
                self.flwr_serverapp_subprocess.wait(timeout=5)
            except subprocess.TimeoutExpired:
                self.flwr_serverapp_subprocess.kill()

    def _build_flwr_run_command(self) -> list[str]:
        """
        Construct the `flwr run` command to execute the Flower application.

        Returns:
            list[str]: A list of command-line arguments for running the Flower application.
        """
        federation_name = self.flwr_run_params.get("federation_name")
        flwr_app_name = self.flwr_run_params.get("flwr_app_name")

        os.environ["TMPDIR"] = os.environ["FLWR_HOME"]

        if self.flwr_run_params.get("patch"):
            command = ["python", "src/patch/flwr_run_patch.py", "run", f"./src/{flwr_app_name}"]
        else:
            command = ["flwr", "run", f"./src/{flwr_app_name}"]

        if federation_name:
            command.append(federation_name)

        return command

    def start(self):
        """Launch the `flower-superlink` and `flwr run` subprocesses using the constructed commands."""
        if self._process is None:
            logger.info(f"[OpenFL Connector] Starting server process: {' '.join(self.flwr_superlink_command)}")
            self._process = subprocess.Popen(self.flwr_superlink_command)
            logger.info(f"[OpenFL Connector] Server process started with PID: {self._process.pid}")
        else:
            logger.info("[OpenFL Connector] Server process is already running.")

        if hasattr(self, 'flwr_run_command') and self.flwr_run_command:
            logger.info(f"[OpenFL Connector] Starting `flwr run` subprocess: {' '.join(self.flwr_run_command)}")
            subprocess.run(self.flwr_run_command)

        if hasattr(self, 'flwr_serverapp_command') and self.flwr_serverapp_command:
            self.interop_client.set_is_flwr_serverapp_running_callback(self.is_flwr_serverapp_running)
            self.flwr_serverapp_subprocess = subprocess.Popen(self.flwr_serverapp_command)

    def stop(self):
        """Terminate the `flower-superlink` subprocess and any associated processes."""
        self._stop_flwr_serverapp()
        if self._process:
            try:
                logger.info(f"[OpenFL Connector] Stopping server process with PID: {self._process.pid}...")
                main_process = psutil.Process(self._process.pid)
                sub_processes = main_process.children(recursive=True)
                for sub_process in sub_processes:
                    logger.info(f"[OpenFL Connector] Stopping server subprocess with PID: {sub_process.pid}...")
                    sub_process.terminate()
                _, still_alive = psutil.wait_procs(sub_processes, timeout=1)
                for p in still_alive:
                    p.kill()
                try:
                    self._process.terminate()
                    self._process.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    self._process.kill()
                self._process = None
                logger.info("[OpenFL Connector] Server process stopped.")
            except Exception as e:
                logger.debug(f"[OpenFL Connector] Error during graceful shutdown: {e}")
                self._process.kill()
                logger.info("[OpenFL Connector] Server process forcefully terminated.")
        else:
            logger.info("[OpenFL Connector] No server process is currently running.")

    def _handle_sigint(self, signum, frame):
        """Handle the SIGINT signal (Ctrl+C) to cleanly stop the server process and its children."""
        logger.info("[OpenFL Connector] SIGINT received. Terminating server process...")
        self.stop()
        sys.exit(0)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Aggregator module."""

import json
import logging
import queue
import time
from threading import Lock
from typing import List, Optional

import openfl.callbacks as callbacks_module
from openfl.component.aggregator.straggler_handling import StragglerPolicy, WaitForAllPolicy
from openfl.databases import PersistentTensorDB, TensorDB
from openfl.interface.aggregation_functions import SecureWeightedAverage, WeightedAverage
from openfl.pipelines import NoCompressionPipeline, TensorCodec
from openfl.protocols import base_pb2, utils
from openfl.protocols.base_pb2 import NamedTensor
from openfl.utilities import TaskResultKey, TensorKey, change_tags

logger = logging.getLogger(__name__)


class Aggregator:
    """An Aggregator is the central node in federated learning.

    Attributes:
        round_number (int): Current round number.
        single_col_cert_common_name (str): Common name for single
            collaborator certificate.
        straggler_handling_policy: Policy for handling stragglers.
        _end_of_round_check_done (list of bool): Indicates if end of round
            check is done for each round.
        stragglers (list): List of stragglers.
        rounds_to_train (int): Number of rounds to train.
        authorized_cols (list of str): IDs of enrolled collaborators.
        uuid (int): Aggregator UUID.
        federation_uuid (str): Federation UUID.
        assigner: Object assigning tasks to collaborators.
        quit_job_sent_to (list): Collaborators sent a quit job.
        tensor_db (TensorDB): Object for tensor database.
        db_store_rounds* (int): Rounds to store in TensorDB.
        logger: Object for logging.
        write_logs (bool): Flag to enable metric writer callback.
        best_model_score (optional): Score of the best model. Defaults to
            None.
        metric_queue (queue.Queue): Queue for metrics.
        compression_pipeline: Pipeline for compressing data.
        tensor_codec (TensorCodec): Codec for tensor compression.
        init_state_path* (str): Initial weight file location.
        best_state_path* (str): Where to store the best model weight.
        last_state_path* (str): Where to store the latest model weight.
        best_tensor_dict (dict): Dict of the best tensors.
        last_tensor_dict (dict): Dict of the last tensors.
        collaborator_tensor_results (dict): Dict of collaborator tensor
            results.
        collaborator_tasks_results (dict): Dict of collaborator tasks
            results.
        collaborator_task_weight (dict): Dict of col task weight.
        lock: A threading Lock object used to ensure thread-safe operations.

    .. note::
        - plan setting
    """

    def __init__(
        self,
        aggregator_uuid,
        federation_uuid,
        authorized_cols,
        init_state_path,
        best_state_path,
        last_state_path,
        assigner,
        use_delta_updates=True,
        straggler_handling_policy: StragglerPolicy = WaitForAllPolicy,
        rounds_to_train=256,
        single_col_cert_common_name=None,
        compression_pipeline=None,
        db_store_rounds=1,
        initial_tensor_dict=None,
        log_memory_usage=False,
        write_logs=False,
        callbacks: Optional[List] = [],
        persist_checkpoint=True,
        persistent_db_path=None,
        secure_aggregation=False,
    ):
        """Initializes the Aggregator.

        Args:
            aggregator_uuid (int): Aggregation ID.
            federation_uuid (str): Federation ID.
            authorized_cols (list of str): The list of IDs of enrolled
                collaborators.
            init_state_path (str): The location of the initial weight file.
            best_state_path (str): The file location to store the weight of
                the best model.
            last_state_path (str): The file location to store the latest
                weight.
            assigner: Assigner object.
            straggler_handling_policy (optional): Straggler handling policy.
            rounds_to_train (int, optional): Number of rounds to train.
                Defaults to 256.
            single_col_cert_common_name (str, optional): Common name for single
                collaborator certificate. Defaults to None.
            compression_pipeline (optional): Compression pipeline. Defaults to
                NoCompressionPipeline.
            db_store_rounds (int, optional): Rounds to store in TensorDB.
                Defaults to 1.
            initial_tensor_dict (dict, optional): Initial tensor dictionary.
            callbacks: List of callbacks to be used during the experiment.
        """
        self.round_number = 0
        self.next_model_round_number = 0

        if single_col_cert_common_name:
            logger.warning(
                "You are running in single collaborator certificate mode. "
                "This mode is intended for development settings only and does not "
                "provide proper Public Key Infrastructure (PKI) security. "
                "Please use this mode with caution."
            )
        # FIXME: "" instead of None is for protobuf compatibility.
        self.single_col_cert_common_name = single_col_cert_common_name or ""

        self.straggler_handling_policy = straggler_handling_policy()

        self.rounds_to_train = rounds_to_train
        self.assigner = assigner
        if self.assigner.is_task_group_evaluation():
            self.rounds_to_train = 1
            logger.info(f"For evaluation tasks setting rounds_to_train = {self.rounds_to_train}")

        self._end_of_round_check_done = [False] * rounds_to_train
        self.stragglers = []

        # if the collaborator requests a delta, this value is set to true
        self.authorized_cols = authorized_cols
        self.uuid = aggregator_uuid
        self.federation_uuid = federation_uuid

        self.quit_job_sent_to = []

        self.tensor_db = TensorDB()
        if persist_checkpoint:
            persistent_db_path = persistent_db_path or "tensor.db"
            logger.info(
                "Persistent checkpoint is enabled, setting persistent db at path %s",
                persistent_db_path,
            )
            self.persistent_db = PersistentTensorDB(persistent_db_path)
        else:
            logger.info("Persistent checkpoint is disabled")
            self.persistent_db = None
        # FIXME: I think next line generates an error on the second round
        # if it is set to 1 for the aggregator.
        self.db_store_rounds = db_store_rounds

        self.best_model_score = None
        self.metric_queue = queue.Queue()

        self.compression_pipeline = compression_pipeline or NoCompressionPipeline()
        self.tensor_codec = TensorCodec(self.compression_pipeline)

        self.init_state_path = init_state_path
        self.best_state_path = best_state_path
        self.last_state_path = last_state_path

        # these enable getting all tensors for a task
        self.collaborator_tasks_results = {}  # {TaskResultKey: list of TensorKeys}
        self.collaborator_task_weight = {}  # {TaskResultKey: data_size}

        # maintain a list of collaborators that have completed task and
        # reported results in a given round
        self.collaborators_done = []
        # Initialize a lock for thread safety
        self.lock = Lock()
        self.use_delta_updates = use_delta_updates

        self.model = None  # Initialize the model attribute to None

        # Callbacks
        self.callbacks = callbacks_module.CallbackList(
            callbacks,
            add_memory_profiler=log_memory_usage,
            add_metric_writer=write_logs,
            origin="aggregator",
        )

        if initial_tensor_dict:
            self._load_initial_tensors_from_dict(initial_tensor_dict)
            self.model = utils.construct_model_proto(
                tensor_dict=initial_tensor_dict,
                round_number=0,
                tensor_pipe=self.compression_pipeline,
            )
        else:
            self.model: base_pb2.ModelProto = utils.load_proto(self.init_state_path)
            self._load_initial_tensors()  # keys are TensorKeys

        self._secure_aggregation_enabled = secure_aggregation
        if self._secure_aggregation_enabled:
            from openfl.utilities.secagg.bootstrap import SecAggSetup

            self.secagg = SecAggSetup(self.uuid, self.authorized_cols, self.tensor_db)

        if self.persistent_db and self._recover():
            logger.info("Recovered state of aggregator")

        # TODO: Aggregator has no concrete notion of round_begin.
        # https://github.com/securefederatedai/openfl/pull/1195#discussion_r1879479537
        self.callbacks.on_experiment_begin()
        self.callbacks.on_round_begin(self.round_number)

    def _recover(self):
        """Populates the aggregator state to the state it was prior a restart"""
        recovered = False
        # load tensors persistent DB
        tensor_key_dict = self.persistent_db.load_tensors(
            self.persistent_db.get_tensors_table_name()
        )
        if len(tensor_key_dict) > 0:
            logger.info(f"Recovering {len(tensor_key_dict)} model tensors")
            recovered = True
            self.tensor_db.cache_tensor(tensor_key_dict)
            committed_round_number, self.best_model_score = (
                self.persistent_db.get_round_and_best_score()
            )
            logger.info("Recovery - Setting model proto")
            to_proto_tensor_dict = {}
            for tk in tensor_key_dict:
                tk_name, _, _, _, _ = tk
                to_proto_tensor_dict[tk_name] = tensor_key_dict[tk]
            self.model = utils.construct_model_proto(
                to_proto_tensor_dict, committed_round_number, self.compression_pipeline
            )
            # round number is the current round which is still in process
            #  i.e. committed_round_number + 1
            self.round_number = committed_round_number + 1
            logger.info(
                "Recovery - loaded round number %s and best score %s",
                self.round_number,
                self.best_model_score,
            )

        next_round_tensor_key_dict = self.persistent_db.load_tensors(
            self.persistent_db.get_next_round_tensors_table_name()
        )
        if len(next_round_tensor_key_dict) > 0:
            logger.info(f"Recovering {len(next_round_tensor_key_dict)} next round model tensors")
            recovered = True
            self.tensor_db.cache_tensor(next_round_tensor_key_dict)

        logger.debug("Recovery - this is the tensor_db after recovery: %s", self.tensor_db)

        if self.persistent_db.is_task_table_empty():
            logger.debug("task table is empty")
            return recovered

        logger.info("Recovery - Replaying saved task results")
        task_id = 1
        while True:
            task_result = self.persistent_db.get_task_result_by_id(task_id)
            if not task_result:
                break
            recovered = True
            collaborator_name = task_result["collaborator_name"]
            round_number = task_result["round_number"]
            task_name = task_result["task_name"]
            data_size = task_result["data_size"]
            serialized_tensors = task_result["named_tensors"]
            named_tensors = [
                NamedTensor.FromString(serialized_tensor)
                for serialized_tensor in serialized_tensors
            ]
            logger.info(
                "Recovery - Replaying task results %s %s %s",
                collaborator_name,
                round_number,
                task_name,
            )
            self.process_task_results(
                collaborator_name, round_number, task_name, data_size, named_tensors
            )
            task_id += 1
        return recovered

    def _load_initial_tensors(self):
        """Load all of the tensors required to begin federated learning.

        Required tensors are: \
            1. Initial model.

        Returns:
            None
        """
        tensor_dict, round_number = utils.deconstruct_model_proto(
            self.model, compression_pipeline=self.compression_pipeline
        )

        # Check selected task_group before updating round number
        if self.assigner.is_task_group_evaluation():
            logger.info("Skipping round_number check for evaluation run")
        elif round_number > self.round_number:
            logger.info(f"Starting training from round {round_number} of previously saved model")
            self.round_number = round_number

        tensor_key_dict = {
            TensorKey(k, self.uuid, self.round_number, False, ("model",)): v
            for k, v in tensor_dict.items()
        }
        # all initial model tensors are loaded here
        self.tensor_db.cache_tensor(tensor_key_dict)
        logger.debug("This is the initial tensor_db: %s", self.tensor_db)

    def _load_initial_tensors_from_dict(self, tensor_dict):
        """Load all of the tensors required to begin federated learning.

        Required tensors are: \
            1. Initial model.

        Returns:
            None
        """
        tensor_key_dict = {
            TensorKey(k, self.uuid, self.round_number, False, ("model",)): v
            for k, v in tensor_dict.items()
        }
        # all initial model tensors are loaded here
        self.tensor_db.cache_tensor(tensor_key_dict)
        logger.debug("This is the initial tensor_db: %s", self.tensor_db)

    def _save_model(self, round_number, file_path):
        """Save the best or latest model.

        Args:
            round_number (int): Model round to be saved.
            file_path (str): Either the best model or latest model file path.

        Returns:
            None
        """
        # Extract the model from TensorDB and set it to the new model
        og_tensor_dict, _ = utils.deconstruct_model_proto(
            self.model, compression_pipeline=self.compression_pipeline
        )
        tensor_keys = [
            TensorKey(k, self.uuid, round_number, False, ("model",))
            for k, v in og_tensor_dict.items()
        ]
        tensor_dict = {}
        tensor_tuple_dict = {}
        next_round_tensors = {}
        for tk in tensor_keys:
            tk_name, _, _, _, _ = tk
            tensor_value = self.tensor_db.get_tensor_from_cache(tk)
            tensor_dict[tk_name] = tensor_value
            tensor_tuple_dict[tk] = tensor_value
            if tensor_dict[tk_name] is None:
                logger.info(
                    "Cannot save model for round %s. Continuing...",
                    round_number,
                )
                return
        if file_path == self.best_state_path:
            self.best_tensor_dict = tensor_dict
        if file_path == self.last_state_path:
            # Transaction to persist/delete all data needed to increment the round
            if self.persistent_db:
                if self.next_model_round_number > 0:
                    next_round_tensors = self.tensor_db.get_tensors_by_round_and_tags(
                        self.next_model_round_number, ("model",)
                    )
                self.persistent_db.finalize_round(
                    tensor_tuple_dict, next_round_tensors, round_number, self.best_model_score
                )
                logger.info(
                    "Persist model and clean task result for round %s",
                    round_number,
                )
            self.last_tensor_dict = tensor_dict
        self.model = utils.construct_model_proto(
            tensor_dict, round_number, self.compression_pipeline
        )
        utils.dump_proto(self.model, file_path)

    def valid_collaborator_cn_and_id(self, cert_common_name, collaborator_common_name):
        """
        Determine if the collaborator certificate and ID are valid for this federation.

        Args:
            cert_common_name (str): Common name for security certificate.
            collaborator_common_name (str): Common name for collaborator.

        Returns:
            bool: True means the collaborator common name matches the name in
                the security certificate.
        """
        # if self.test_mode_whitelist is None, then the common_name must
        # match collaborator_common_name and be in authorized_cols
        # FIXME: '' instead of None is just for protobuf compatibility.
        #  Cleaner solution?
        if self.single_col_cert_common_name == "":
            return (
                cert_common_name == collaborator_common_name
                and collaborator_common_name in self.authorized_cols
            )
        # otherwise, common_name must be in whitelist and
        # collaborator_common_name must be in authorized_cols
        else:
            return (
                cert_common_name == self.single_col_cert_common_name
                and collaborator_common_name in self.authorized_cols
            )

    def all_quit_jobs_sent(self):
        """Assert all quit jobs are sent to collaborators.

        Returns:
            bool: True if all quit jobs are sent, False otherwise.
        """
        return set(self.quit_job_sent_to) == set(self.authorized_cols)

    @staticmethod
    def _get_sleep_time():
        """Sleep 10 seconds.

        Returns:
            int: Sleep time.
        """
        # Decrease sleep period for finer discretezation
        return 10

    def _time_to_quit(self):
        """If all rounds are complete, it's time to quit.

        Returns:
            bool: True if it's time to quit, False otherwise.
        """
        return self.round_number >= self.rounds_to_train

    def get_tasks(self, collaborator_name):
        """RPC called by a collaborator to determine which tasks to perform.

        Args:
            collaborator_name (str): Requested collaborator name.

        Returns:
            tasks (list[str]): List of tasks to be performed by the requesting
                collaborator for the current round.
            round_number (int): Actual round number.
            sleep_time (int): Sleep time.
            time_to_quit (bool): Whether it's time to quit.
        """
        logger.debug(
            f"Aggregator GetTasks function reached from collaborator {collaborator_name}..."
        )

        # first, if it is time to quit, inform the collaborator
        if self._time_to_quit():
            logger.info(
                "Sending signal to collaborator %s to shutdown...",
                collaborator_name,
            )
            self.quit_job_sent_to.append(collaborator_name)

            tasks = None
            sleep_time = 0
            time_to_quit = True

            return tasks, self.round_number, sleep_time, time_to_quit

        time_to_quit = False
        # otherwise, get the tasks from our task assigner
        tasks = self.assigner.get_tasks_for_collaborator(collaborator_name, self.round_number)

        # if no tasks, tell the collaborator to sleep
        if len(tasks) == 0:
            tasks = None
            sleep_time = Aggregator._get_sleep_time()

            return tasks, self.round_number, sleep_time, time_to_quit

        # if we do have tasks, remove any that we already have results for
        if isinstance(tasks[0], str):
            # backward compatibility
            tasks = [
                t
                for t in tasks
                if not self._collaborator_task_completed(collaborator_name, t, self.round_number)
            ]
            if collaborator_name in self.stragglers:
                tasks = []

        else:
            tasks = [
                t
                for t in tasks
                if not self._collaborator_task_completed(
                    collaborator_name, t.name, self.round_number
                )
            ]
            if collaborator_name in self.stragglers:
                tasks = []

        # Do the check again because it's possible that all tasks have
        # been completed
        if len(tasks) == 0:
            tasks = None
            sleep_time = Aggregator._get_sleep_time()

            return tasks, self.round_number, sleep_time, time_to_quit

        logger.info(
            f"Sending tasks to collaborator {collaborator_name} for round {self.round_number}"
        )
        sleep_time = 0

        # Start straggler handling policy for timer based callback is required
        # for %age based policy callback is not required
        self.straggler_handling_policy.start_policy(callback=self._straggler_cutoff_time_elapsed)

        return tasks, self.round_number, sleep_time, time_to_quit

    def _straggler_cutoff_time_elapsed(self) -> None:
        """
        This method is called by the straggler handling policy when cutoff timer is elapsed.
        It applies straggler handling policy and ends the round early.

        Returns:
            None
        """
        logger.warning(
            f"Round number: {self.round_number} cutoff timer elapsed after "
            f"{self.straggler_handling_policy.straggler_cutoff_time}s. "
            f"Applying {self.straggler_handling_policy.__class__.__name__} policy."
        )

        with self.lock:
            # Check if minimum collaborators reported results
            self._end_of_round_with_stragglers_check()

    def get_aggregated_tensor(
        self,
        tensor_name,
        round_number,
        report,
        tags,
        require_lossless,
    ):
        """
        RPC called by collaborator.

        Performs local lookup to determine if there is an aggregated tensor available
        that matches the request.

        Args:
            tensor_name (str): Name of the tensor.
            round_number (int): Actual round number.
            report (bool): Whether to report.
            tags (tuple[str, ...]): Tags.
            require_lossless (bool): Whether to require lossless.

        Returns:
            named_tensor (protobuf) :  NamedTensor, the tensor requested by the collaborator.

        Raises:
            ValueError: if Aggregator does not have an aggregated tensor for {tensor_key}.
        """
        if "compressed" in tags or require_lossless:
            compress_lossless = True
        else:
            compress_lossless = False

        # TODO the TensorDB doesn't support compressed data yet.
        #  The returned tensor will
        # be recompressed anyway.
        if "compressed" in tags:
            tags = change_tags(tags, remove_field="compressed")
        if "lossy_compressed" in tags:
            tags = change_tags(tags, remove_field="lossy_compressed")

        tensor_key = TensorKey(tensor_name, self.uuid, round_number, report, tags)
        tensor_name, origin, round_number, report, tags = tensor_key

        if "aggregated" in tags and "delta" in tags and round_number != 0:
            agg_tensor_key = TensorKey(tensor_name, origin, round_number, report, ("aggregated",))
        else:
            agg_tensor_key = tensor_key

        nparray = self.tensor_db.get_tensor_from_cache(agg_tensor_key)

        start_retrieving_time = time.time()
        while nparray is None:
            logger.debug("Waiting for tensor_key %s", agg_tensor_key)
            time.sleep(5)
            nparray = self.tensor_db.get_tensor_from_cache(agg_tensor_key)
            if (time.time() - start_retrieving_time) > 60:
                break

        if nparray is None:
            raise ValueError(f"Aggregator does not have an aggregated tensor for {tensor_key}")

        # quite a bit happens in here, including compression, delta handling,
        # etc...
        # we might want to cache these as well
        named_tensor = self._nparray_to_named_tensor(
            agg_tensor_key, nparray, send_model_deltas=True, compress_lossless=compress_lossless
        )

        return named_tensor

    def _nparray_to_named_tensor(self, tensor_key, nparray, send_model_deltas, compress_lossless):
        """Construct the NamedTensor Protobuf.

        Also includes logic to create delta, compress tensors with the
            TensorCodec, etc.

        Args:
            tensor_key (TensorKey): Tensor key.
            nparray (np.array): Numpy array.
            send_model_deltas (bool): Whether to send model deltas.
            compress_lossless (bool): Whether to compress lossless.

        Returns:
            tensor_key (TensorKey): Tensor key.
            nparray (np.array): Numpy array.

        """
        tensor_name, origin, round_number, report, tags = tensor_key
        # Secure aggregation setup tensor.
        if "secagg" in tags:
            import numpy as np

            class NumpyEncoder(json.JSONEncoder):
                def default(self, obj):
                    if isinstance(obj, np.ndarray):
                        return obj.tolist()
                    return super().default(obj)

            named_tensor = utils.construct_named_tensor(
                tensor_key,
                str.encode(json.dumps(nparray, cls=NumpyEncoder)),
                {},
                lossless=True,
            )

            return named_tensor
        # if we have an aggregated tensor, we can make a delta
        if "aggregated" in tags and send_model_deltas:
            # Should get the pretrained model to create the delta. If training
            # has happened, Model should already be stored in the TensorDB
            model_tk = TensorKey(tensor_name, origin, round_number - 1, report, ("model",))

            model_nparray = self.tensor_db.get_tensor_from_cache(model_tk)

            assert model_nparray is not None, (
                "The original model layer should be present if the latest "
                "aggregated model is present"
            )
            delta_tensor_key, delta_nparray = self.tensor_codec.generate_delta(
                tensor_key, nparray, model_nparray
            )
            delta_comp_tensor_key, delta_comp_nparray, metadata = self.tensor_codec.compress(
                delta_tensor_key, delta_nparray, lossless=compress_lossless
            )
            named_tensor = utils.construct_named_tensor(
                delta_comp_tensor_key,
                delta_comp_nparray,
                metadata,
                lossless=compress_lossless,
            )

        else:
            # Assume every other tensor requires lossless compression
            compressed_tensor_key, compressed_nparray, metadata = self.tensor_codec.compress(
                tensor_key, nparray, require_lossless=True
            )
            named_tensor = utils.construct_named_tensor(
                compressed_tensor_key,
                compressed_nparray,
                metadata,
                lossless=compress_lossless,
            )

        return named_tensor

    def _collaborator_task_completed(self, collaborator, task_name, round_num):
        """Check if the collaborator has completed the task for the round.

         The aggregator doesn't actually know which tensors should be sent from
         the collaborator so it must to rely specifically on the presence of
         previous results.

        Args:
         collaborator (str): Collaborator to check if their task has been
             completed.
         task_name (str): The name of the task (TaskRunner function).
         round_num (int): Round number.

         Returns:
             bool: Whether or not the collaborator has completed the task for
                 this round.
        """
        task_key = TaskResultKey(task_name, collaborator, round_num)
        return task_key in self.collaborator_tasks_results

    def send_local_task_results(
        self,
        collaborator_name,
        round_number,
        task_name,
        data_size,
        named_tensors,
    ):
        """
        RPC called by collaborator.

        Transmits collaborator's task results to the aggregator.

        Args:
            collaborator_name (str): Collaborator name.
            round_number (int): Round number.
            task_name (str): Task name.
            data_size (int): Data size.
            named_tensors (protobuf NamedTensor): Named tensors.

        Returns:
            None
        """
        # Check if secure aggregation is enabled.
        if self._secure_aggregation_enabled:
            secagg_setup = self.secagg.process_secagg_setup_tensors(named_tensors)
            # Task results processing is not required if the tensors belong to
            # secure aggregation setup stage.
            if secagg_setup:
                return

        if self.persistent_db:
            # Save task and its metadata for recovery
            serialized_tensors = [tensor.SerializeToString() for tensor in named_tensors]
            self.persistent_db.save_task_results(
                collaborator_name, round_number, task_name, data_size, serialized_tensors
            )
            logger.debug(
                f"Persisting task results {task_name} from {collaborator_name} round {round_number}"
            )
        logger.info(
            f"Collaborator {collaborator_name} is sending task results "
            f"for {task_name}, round {round_number}"
        )

        self.process_task_results(
            collaborator_name, round_number, task_name, data_size, named_tensors
        )

    def process_task_results(
        self,
        collaborator_name,
        round_number,
        task_name,
        data_size,
        named_tensors,
    ):
        if self._time_to_quit() or collaborator_name in self.stragglers:
            logger.warning(
                f"STRAGGLER: Collaborator {collaborator_name} is reporting results "
                f"after task {task_name} has finished."
            )
            return

        if self.round_number != round_number:
            logger.warning(
                f"Collaborator {collaborator_name} is reporting results"
                f" for the wrong round: {round_number}. Ignoring..."
            )
            return

        task_key = TaskResultKey(task_name, collaborator_name, round_number)

        # we mustn't have results already
        if self._collaborator_task_completed(collaborator_name, task_name, round_number):
            logger.warning(
                f"Aggregator already has task results from collaborator {collaborator_name}"
                f" for task {task_key}"
            )
            return

        # By giving task_key it's own weight, we can support different
        # training/validation weights
        # As well as eventually supporting weights that change by round
        # (if more data is added)
        self.collaborator_task_weight[task_key] = data_size

        # initialize the list of tensors that go with this task
        # Setting these incrementally is leading to missing values
        task_results = []

        for named_tensor in named_tensors:
            # quite a bit happens in here, including decompression, delta
            # handling, etc...
            tensor_key, value = self._process_named_tensor(named_tensor, collaborator_name)

            if "metric" in tensor_key.tags:
                # Caution: This schema must be followed. It is also used in
                # gRPC message streams for director/envoy.
                metrics = {
                    "round": round_number,
                    "metric_origin": collaborator_name,
                    "task_name": task_name,
                    "metric_name": tensor_key.tensor_name,
                    "metric_value": float(value),
                }
                self.metric_queue.put(metrics)

            task_results.append(tensor_key)

        self.collaborator_tasks_results[task_key] = task_results

        # Check if collaborator or round is done.
        self._is_collaborator_done(collaborator_name, round_number)
        self._end_of_round_with_stragglers_check()

    def _end_of_round_with_stragglers_check(self):
        """
        Checks if the minimum required collaborators have reported their results,
        identifies any stragglers, and initiates an early round end if necessary.

        Returns:
            None
        """
        if self.straggler_handling_policy.straggler_cutoff_check(
            len(self.collaborators_done), len(self.authorized_cols)
        ):
            self.stragglers = [
                collab_name
                for collab_name in self.authorized_cols
                if collab_name not in self.collaborators_done
            ]
            if len(self.stragglers) != 0:
                logger.warning(f"Identified stragglers: {self.stragglers}")
            self._end_of_round_check()

    def _process_named_tensor(self, named_tensor, collaborator_name):
        """Extract the named tensor fields.

        Performs decompression, delta computation, and inserts results into
        TensorDB.

        Args:
            named_tensor (protobuf NamedTensor): Named tensor.
                protobuf that will be extracted from and processed
            collaborator_name (str): Collaborator name.
                Collaborator name is needed for proper tagging of resulting
                tensorkeys.

        Returns:
            tensor_key (TensorKey): Tensor key.
                The tensorkey extracted from the protobuf.
            nparray (np.array): Numpy array.
                The numpy array associated with the returned tensorkey.
        """
        raw_bytes = named_tensor.data_bytes
        metadata = [
            {
                "int_to_float": proto.int_to_float,
                "int_list": proto.int_list,
                "bool_list": proto.bool_list,
            }
            for proto in named_tensor.transformer_metadata
        ]
        # The tensor has already been transferred to aggregator,
        # so the newly constructed tensor should have the aggregator origin
        tensor_key = TensorKey(
            named_tensor.name,
            self.uuid,
            named_tensor.round_number,
            named_tensor.report,
            tuple(named_tensor.tags),
        )
        tensor_name, origin, round_number, report, tags = tensor_key

        assert "compressed" in tags or "lossy_compressed" in tags, (
            f"Named tensor {tensor_key} is not compressed"
        )
        if "compressed" in tags:
            dec_tk, decompressed_nparray = self.tensor_codec.decompress(
                tensor_key,
                data=raw_bytes,
                transformer_metadata=metadata,
                require_lossless=True,
            )
            dec_name, dec_origin, dec_round_num, dec_report, dec_tags = dec_tk
            # Need to add the collaborator tag to the resulting tensor
            new_tags = change_tags(dec_tags, add_field=collaborator_name)

            # layer.agg.n.trained.delta.col_i
            decompressed_tensor_key = TensorKey(
                dec_name, dec_origin, dec_round_num, dec_report, new_tags
            )
        if "lossy_compressed" in tags:
            dec_tk, decompressed_nparray = self.tensor_codec.decompress(
                tensor_key,
                data=raw_bytes,
                transformer_metadata=metadata,
                require_lossless=False,
            )
            dec_name, dec_origin, dec_round_num, dec_report, dec_tags = dec_tk
            new_tags = change_tags(dec_tags, add_field=collaborator_name)
            # layer.agg.n.trained.delta.lossy_decompressed.col_i
            decompressed_tensor_key = TensorKey(
                dec_name, dec_origin, dec_round_num, dec_report, new_tags
            )

        if "delta" in tags:
            base_model_tensor_key = TensorKey(tensor_name, origin, round_number, report, ("model",))
            base_model_nparray = self.tensor_db.get_tensor_from_cache(base_model_tensor_key)
            if base_model_nparray is None:
                raise ValueError(f"Base model {base_model_tensor_key} not present in TensorDB")
            final_tensor_key, final_nparray = self.tensor_codec.apply_delta(
                decompressed_tensor_key,
                decompressed_nparray,
                base_model_nparray,
            )
        else:
            final_tensor_key = decompressed_tensor_key
            final_nparray = decompressed_nparray

        assert final_nparray is not None, f"Could not create tensorkey {final_tensor_key}"
        self.tensor_db.cache_tensor({final_tensor_key: final_nparray})
        logger.debug("Created TensorKey: %s", final_tensor_key)

        return final_tensor_key, final_nparray

    def _prepare_trained(self, tensor_name, origin, round_number, report, agg_results):
        """Prepare aggregated tensorkey tags.

        Args:
            tensor_name (str): Tensor name.
            origin: Origin.
            round_number (int): Round number.
            report (bool): Whether to report.
            agg_results (np.array): Aggregated results.
        """
        # The aggregated tensorkey tags should have the form of
        # 'trained' or 'trained.lossy_decompressed'
        # They need to be relabeled to 'aggregated' and
        # reinserted. Then delta performed, compressed, etc.
        # then reinserted to TensorDB with 'model' tag

        # First insert the aggregated model layer with the
        # correct tensorkey
        agg_tag_tk = TensorKey(tensor_name, origin, round_number + 1, report, ("aggregated",))
        self.tensor_db.cache_tensor({agg_tag_tk: agg_results})

        # Create delta and save it in TensorDB
        base_model_tk = TensorKey(tensor_name, origin, round_number, report, ("model",))
        base_model_nparray = self.tensor_db.get_tensor_from_cache(base_model_tk)
        if base_model_nparray is not None and self.use_delta_updates:
            delta_tk, delta_nparray = self.tensor_codec.generate_delta(
                agg_tag_tk, agg_results, base_model_nparray
            )
        else:
            # This condition is possible for base model
            # optimizer states (i.e. Adam/iter:0, SGD, etc.)
            # These values couldn't be present for the base
            # model because no training occurs on the aggregator
            delta_tk, delta_nparray = agg_tag_tk, agg_results

        # Compress lossless/lossy
        compressed_delta_tk, compressed_delta_nparray, metadata = self.tensor_codec.compress(
            delta_tk, delta_nparray
        )

        # TODO extend the TensorDB so that compressed data is
        #  supported. Once that is in place
        # the compressed delta can just be stored here instead
        # of recreating it for every request

        # Decompress lossless/lossy
        decompressed_delta_tk, decompressed_delta_nparray = self.tensor_codec.decompress(
            compressed_delta_tk, compressed_delta_nparray, metadata
        )

        self.tensor_db.cache_tensor({decompressed_delta_tk: decompressed_delta_nparray})

        # Apply delta (unless delta couldn't be created)
        if base_model_nparray is not None and self.use_delta_updates:
            logger.debug("Applying delta for layer %s", decompressed_delta_tk[0])
            new_model_tk, new_model_nparray = self.tensor_codec.apply_delta(
                decompressed_delta_tk,
                decompressed_delta_nparray,
                base_model_nparray,
            )
        else:
            new_model_tk, new_model_nparray = (
                decompressed_delta_tk,
                decompressed_delta_nparray,
            )

        # Now that the model has been compressed/decompressed
        # with delta operations,
        # Relabel the tags to 'model'
        (
            new_model_tensor_name,
            new_model_origin,
            new_model_round_number,
            new_model_report,
            new_model_tags,
        ) = new_model_tk
        final_model_tk = TensorKey(
            new_model_tensor_name,
            new_model_origin,
            new_model_round_number,
            new_model_report,
            ("model",),
        )
        self.next_model_round_number = new_model_round_number
        # Finally, cache the updated model tensor
        self.tensor_db.cache_tensor({final_model_tk: new_model_nparray})

    def _compute_validation_related_task_metrics(self, task_name) -> dict:
        """Compute all validation related metrics.

        Args:
            task_name (str): Task name.

        Returns:
            A dictionary of reportable metrics.
        """
        # By default, print out all of the metrics that the validation
        # task sent
        # This handles getting the subset of collaborators that may be
        # part of the validation task
        all_collaborators_for_task = self.assigner.get_collaborators_for_task(
            task_name, self.round_number
        )
        # Leave out straggler for the round even if they've partially
        # completed given tasks
        collaborators_for_task = []
        collaborators_for_task = [
            c for c in all_collaborators_for_task if c in self.collaborators_done
        ]

        # The collaborator data sizes for that task
        collaborator_weights_unnormalized = {
            c: self.collaborator_task_weight[TaskResultKey(task_name, c, self.round_number)]
            for c in collaborators_for_task
        }
        weight_total = sum(collaborator_weights_unnormalized.values())
        collaborator_weight_dict = {
            k: v / weight_total for k, v in collaborator_weights_unnormalized.items()
        }

        # The validation task should have just a couple tensors (i.e.
        # metrics) associated with it. Because each collaborator should
        # have sent the same tensor list, we can use the first
        # collaborator in our subset, and apply the correct
        # transformations to the tensorkey to resolve the aggregated
        # tensor for that round
        task_agg_function = self.assigner.get_aggregation_type_for_task(task_name)
        task_key = TaskResultKey(task_name, collaborators_for_task[0], self.round_number)

        metrics = {}
        for tensor_key in self.collaborator_tasks_results[task_key]:
            tensor_name, origin, round_number, report, tags = tensor_key
            assert collaborators_for_task[0] in tags, (
                f"Tensor {tensor_key} in task {task_name} has not been processed correctly"
            )
            # Strip the collaborator label, and lookup aggregated tensor
            new_tags = change_tags(tags, remove_field=collaborators_for_task[0])
            agg_tensor_key = TensorKey(tensor_name, origin, round_number, report, new_tags)
            # Check if secure aggregation is enabled, set aggregation function.
            agg_function = task_agg_function
            if "metric" in tags:
                agg_function = WeightedAverage()
            elif self._secure_aggregation_enabled:
                agg_function = SecureWeightedAverage()
            agg_results = self.tensor_db.get_aggregated_tensor(
                agg_tensor_key,
                collaborator_weight_dict,
                aggregation_function=agg_function,
            )

            if report:
                # Metric must be a scalar.
                value = float(agg_results)

                # TODO: Deprecate `metric_queue` going forward.
                self.metric_queue.put(
                    {
                        "metric_origin": "aggregator",
                        "task_name": task_name,
                        "metric_name": tensor_key.tensor_name,
                        "metric_value": value,
                        "round": round_number,
                    }
                )
                metrics.update({f"aggregator/{task_name}/{tensor_key.tensor_name}": value})

                # FIXME: Configurable logic for min/max criteria in saving best.
                if "validate_agg" in tags:
                    # Compare the accuracy of the model, potentially save it.
                    if self.best_model_score is None or self.best_model_score < agg_results:
                        logger.info(
                            f"Round {round_number}: saved the best model with score {agg_results:f}"
                        )
                        self.best_model_score = agg_results
                        self._save_model(round_number, self.best_state_path)
            if "trained" in tags:
                self._prepare_trained(tensor_name, origin, round_number, report, agg_results)

        return metrics

    def _end_of_round_check(self):
        """Check if the round complete.

        If so, perform many end of round operations,
        such as model aggregation, metric reporting, delta generation (+
        associated tensorkey labeling), and save the model.

        Args:
            None

        Returns:
            None
        """
        if self._end_of_round_check_done[self.round_number]:
            return

        # Compute all validation related metrics
        logs = {}
        for task_name in self.assigner.get_all_tasks_for_round(self.round_number):
            logs.update(self._compute_validation_related_task_metrics(task_name))

        # End of round callbacks.
        self.callbacks.on_round_end(self.round_number, logs)

        # Once all of the task results have been processed
        self._end_of_round_check_done[self.round_number] = True

        # Save the latest model
        logger.info("Saving round %s model...", self.round_number)
        self._save_model(self.round_number, self.last_state_path)

        self.round_number += 1
        # resetting stragglers for task for a new round
        self.stragglers = []
        # resetting collaborators_done for next round
        self.collaborators_done = []
        self.collaborator_tasks_results = {}
        self.collaborator_task_weight = {}

        # TODO This needs to be fixed!
        if self._time_to_quit():
            logger.info("Experiment Completed. Cleaning up...")
        else:
            logger.info("Starting round %s...", self.round_number)
            # https://github.com/securefederatedai/openfl/pull/1195#discussion_r1879479537
            self.callbacks.on_round_begin(self.round_number)

        # Cleaning tensor db
        self.tensor_db.clean_up(self.db_store_rounds)
        # Reset straggler handling policy for the next round.
        self.straggler_handling_policy.reset_policy_for_round()

    def _is_collaborator_done(self, collaborator_name: str, round_number: int) -> None:
        """
        Check if all tasks given to the collaborator are completed then,
        completed or not.

        Args:
            collaborator_name (str): Collaborator name.
            round_number (int): Round number.

        Returns:
            None
        """
        if self.round_number != round_number:
            logger.warning(
                f"Collaborator {collaborator_name} is reporting results"
                f" for the wrong round: {round_number}. Ignoring..."
            )
            return

        # Get all tasks given to the collaborator for current round
        all_tasks = self.assigner.get_tasks_for_collaborator(collaborator_name, self.round_number)
        # Check if all given tasks are completed by the collaborator
        all_tasks_completed = True
        for task in all_tasks:
            if hasattr(task, "name"):
                task = task.name
            all_tasks_completed = all_tasks_completed and self._collaborator_task_completed(
                collaborator=collaborator_name, task_name=task, round_num=self.round_number
            )
        # If the collaborator has completed ALL tasks for current round,
        # update collaborators_done
        if all_tasks_completed:
            self.collaborators_done.append(collaborator_name)
            logger.info(
                f"Round {self.round_number}: Collaborators that have completed all tasks: "
                f"{self.collaborators_done}"
            )

    def stop(self, failed_collaborator: str = None) -> None:
        """Stop aggregator execution.

        Args:
            failed_collaborator (str, optional): Failed collaborator. Defaults to None.

        Returns:
            None
        """
        logger.info("Force stopping the aggregator execution.")
        # We imitate quit_job_sent_to the failed collaborator
        # So the experiment set to a finished state
        if failed_collaborator:
            self.quit_job_sent_to.append(failed_collaborator)

        # End of experiment callbacks.
        self.callbacks.on_experiment_end()

        # This code does not actually send `quit` tasks to collaborators,
        # it just mimics it by filling arrays.
        for collaborator_name in filter(lambda c: c != failed_collaborator, self.authorized_cols):
            logger.info(
                "Sending signal to collaborator %s to shutdown...",
                collaborator_name,
            )
            self.quit_job_sent_to.append(collaborator_name)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.utilities package."""

import inspect
import itertools
from types import MethodType

import numpy as np

from openfl.experimental.workflow.utilities import ResourcesAllocationError


def parse_attrs(ctx, exclude=[], reserved_words=["next", "runtime", "input"]):
    """Parses the context to get its attributes and artifacts, excluding those
    specified.

    Args:
        ctx (any): The context to parse.
        exclude (list, optional): A list of attribute names to exclude.
            Defaults to an empty list.
        reserved_words (list, optional): A list of reserved words to exclude.
            Defaults to ["next", "runtime", "input"].

    Returns:
        tuple: A tuple containing a list of attribute names and a list of
            valid artifacts (pairs of attribute names and values).
    """
    # TODO Persist attributes to local disk, database, object store, etc. here
    cls_attrs = []
    valid_artifacts = []
    for i in inspect.getmembers(ctx):
        if (
            not hasattr(i[1], "task")
            and not i[0].startswith("_")
            and i[0] not in reserved_words
            and i[0] not in exclude
            and i not in inspect.getmembers(type(ctx))
        ):
            if not isinstance(i[1], MethodType):
                cls_attrs.append(i[0])
                valid_artifacts.append((i[0], i[1]))
    return cls_attrs, valid_artifacts


def generate_artifacts(ctx, reserved_words=["next", "runtime", "input", "checkpoint"]):
    """Generates artifacts from the given context, excluding specified reserved
    words.

    Args:
        ctx (any): The context to generate artifacts from.
        reserved_words (list, optional): A list of reserved words to exclude.
            Defaults to ["next", "runtime", "input", "checkpoint"].

    Returns:
        tuple: A tuple containing a generator of artifacts and a list of
            attribute names.
    """
    cls_attrs, valid_artifacts = parse_attrs(ctx, reserved_words=reserved_words)

    def artifacts_iter():
        # Helper function from metaflow source
        while valid_artifacts:
            var, val = valid_artifacts.pop()
            yield var, val

    return artifacts_iter, cls_attrs


def filter_attributes(ctx, f, **kwargs):
    """Filters out attributes from the next task in the flow based on inclusion
    or exclusion.

    Args:
        ctx (any): The context to filter attributes from.
        f (function): The next task function in the flow.
        **kwargs: Optional arguments that specify the 'include' or 'exclude'
            lists.

    Raises:
        RuntimeError: If both 'include' and 'exclude' are present, or if an
            attribute in 'include' or 'exclude' is not found in the context's
            attributes.
    """
    _, cls_attrs = generate_artifacts(ctx=ctx)

    # Validate input arguments
    _validate_include_exclude(kwargs, cls_attrs)

    # Process the 'include' or 'exclude' lists
    if "include" in kwargs:
        _process_inclusion(ctx, cls_attrs, kwargs["include"], f)
    elif "exclude" in kwargs:
        _process_exclusion(ctx, cls_attrs, kwargs["exclude"], f)


def _validate_include_exclude(kwargs, cls_attrs):
    """Validates that 'include' and 'exclude' are not both present, and that
    attributes in 'include' or 'exclude' exist in the context.

    Args:
        kwargs (dict): The keyword arguments passed to filter_attributes.
        cls_attrs (list): The list of all attributes in the context.

    Raises:
        RuntimeError: If both 'include' and 'exclude' are present, or if any
            attribute in 'include' or 'exclude' does not exist in the context.
    """
    if "include" in kwargs and "exclude" in kwargs:
        raise RuntimeError("'include' and 'exclude' should not both be present")

    if "include" in kwargs:
        assert isinstance(kwargs["include"], list)
        for attr in kwargs["include"]:
            if attr not in cls_attrs:
                raise RuntimeError(f"Argument '{attr}' not found in flow task")

    if "exclude" in kwargs:
        assert isinstance(kwargs["exclude"], list)
        for attr in kwargs["exclude"]:
            if attr not in cls_attrs:
                raise RuntimeError(f"Argument '{attr}' not found in flow task")


def _process_inclusion(ctx, cls_attrs, include_list, f):
    """Handles the inclusion logic: removes attributes not in the 'include' list.

    Args:
        ctx (any): The context to filter attributes from.
        cls_attrs (list): The list of all attributes in the class.
        include_list (list): The list of attributes to keep.
        f (function): The task function to be processed.
    """
    for attr in cls_attrs:
        if attr not in include_list:
            delattr(ctx, attr)


def _process_exclusion(ctx, cls_attrs, exclude_list, f):
    """Handles the exclusion logic: removes attributes in the 'exclude' list.

    Args:
        ctx (any): The context to filter attributes from.
        cls_attrs (list): The list of all attributes in the class.
        exclude_list (list): The list of attributes to exclude.
        f (function): The task function to be processed.
    """
    for attr in cls_attrs:
        if attr in exclude_list and hasattr(ctx, attr):
            delattr(ctx, attr)


def checkpoint(ctx, parent_func, chkpnt_reserved_words=["next", "runtime"]):
    """Optionally saves the current state for the task just executed.

    Args:
        ctx (any): The context to checkpoint.
        parent_func (function): The function that was just executed.
        chkpnt_reserved_words (list, optional): A list of reserved words to
            exclude from checkpointing. Defaults to ["next", "runtime"].

    Returns:
        step_stdout (io.StringIO): parent_func stdout
    """

    # Extract the stdout & stderr from the buffer
    # NOTE: Any prints in this method before this line will be recorded as
    # step output/error
    step_stdout, step_stderr = parent_func._stream_buffer.get_stdstream()

    if ctx._checkpoint:
        # all objects will be serialized using Metaflow interface
        print(f"Saving data artifacts for {parent_func.__name__}")
        artifacts_iter, _ = generate_artifacts(ctx=ctx, reserved_words=chkpnt_reserved_words)
        task_id = ctx._metaflow_interface.create_task(parent_func.__name__)
        ctx._metaflow_interface.save_artifacts(
            artifacts_iter(),
            task_name=parent_func.__name__,
            task_id=task_id,
            buffer_out=step_stdout,
            buffer_err=step_stderr,
        )
        print(f"Saved data artifacts for {parent_func.__name__}")
        return step_stdout


def old_check_resource_allocation(num_gpus, each_participant_gpu_usage):
    remaining_gpu_memory = {}
    # TODO for each GPU the function tries see if all participant usages fit
    # into a GPU, it it doesn't it removes that participant from the
    # participant list, and adds it to the remaining_gpu_memory dict. So any
    # sum of GPU requirements above 1 triggers this.
    # But at this point the function will raise an error because
    # remaining_gpu_memory is never cleared.
    # The participant list should remove the participant if it fits in the gpu
    # and save the partipant if it doesn't and continue to the next GPU to see
    # if it fits in that one, only if we run out of GPUs should this function
    # raise an error.
    for gpu in np.ones(num_gpus, dtype=int):
        for i, (participant_name, participant_gpu_usage) in enumerate(
            each_participant_gpu_usage.items()
        ):
            if gpu == 0:
                break
            if gpu < participant_gpu_usage:
                remaining_gpu_memory.update({participant_name: gpu})
                each_participant_gpu_usage = dict(
                    itertools.islice(each_participant_gpu_usage.items(), i)
                )
            else:
                gpu -= participant_gpu_usage
    if len(remaining_gpu_memory) > 0:
        raise ResourcesAllocationError(
            f"Failed to allocate Participant {list(remaining_gpu_memory.keys())} "
            + "to specified GPU. Please try allocating lesser GPU resources to participants"
        )


def check_resource_allocation(num_gpus, each_participant_gpu_usage):
    # copy participant dict
    need_assigned = each_participant_gpu_usage.copy()
    # cycle through all available GPU availability
    for gpu in np.ones(num_gpus, dtype=int):
        # buffer to cycle though since need_assigned will change sizes as we
        # assign participants
        current_dict = need_assigned.copy()
        for participant_name, participant_gpu_usage in current_dict.items():
            if gpu == 0:
                break
            if gpu < participant_gpu_usage:
                # participant doesn't fitm break to next GPU
                break
            else:
                # if participant fits remove from need_assigned
                need_assigned.pop(participant_name)
                gpu -= participant_gpu_usage

    # raise error if after going though all gpus there are still participants
    # that needed to be assigned
    if len(need_assigned) > 0:
        raise ResourcesAllocationError(
            f"Failed to allocate Participant {list(need_assigned.keys())} "
            + "to specified GPU. Please try allocating lesser GPU resources to participants"
        )



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""AggregatorGRPCClient module."""

import time
from logging import getLogger
from typing import Optional, Tuple

import grpc

from openfl.experimental.workflow.protocols import aggregator_pb2, aggregator_pb2_grpc
from openfl.experimental.workflow.transport.grpc.grpc_channel_options import channel_options


class ConstantBackoff:
    """Constant Backoff policy."""

    def __init__(self, reconnect_interval, logger, uri):
        """Initialize Constant Backoff."""
        self.reconnect_interval = reconnect_interval
        self.logger = logger
        self.uri = uri

    def sleep(self):
        """Sleep for specified interval."""
        self.logger.info(f"Attempting to connect to aggregator at {self.uri}")
        time.sleep(self.reconnect_interval)


class RetryOnRpcErrorClientInterceptor(
    grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor
):
    """Retry gRPC connection on failure."""

    def __init__(
        self,
        sleeping_policy,
        status_for_retry: Optional[Tuple[grpc.StatusCode]] = None,
    ):
        """Initialize function for gRPC retry."""
        self.sleeping_policy = sleeping_policy
        self.status_for_retry = status_for_retry

    def _intercept_call(self, continuation, client_call_details, request_or_iterator):
        """Intercept the call to the gRPC server."""
        while True:
            response = continuation(client_call_details, request_or_iterator)

            if isinstance(response, grpc.RpcError):
                # If status code is not in retryable status codes
                self.sleeping_policy.logger.info(f"Response code: {response.code()}")
                if self.status_for_retry and response.code() not in self.status_for_retry:
                    return response

                self.sleeping_policy.sleep()
            else:
                return response

    def intercept_unary_unary(self, continuation, client_call_details, request):
        """Wrap intercept call for unary->unary RPC."""
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        """Wrap intercept call for stream->unary RPC."""
        return self._intercept_call(continuation, client_call_details, request_iterator)


def _atomic_connection(func):
    def wrapper(self, *args, **kwargs):
        self.reconnect()
        response = func(self, *args, **kwargs)
        self.disconnect()
        return response

    return wrapper


def _resend_data_on_reconnection(func):
    def wrapper(self, *args, **kwargs):
        while True:
            try:
                response = func(self, *args, **kwargs)
            except grpc.RpcError as e:
                if e.code() == grpc.StatusCode.UNKNOWN:
                    self.logger.info(
                        f"Attempting to resend data request to aggregator at {self.uri}"
                    )
                elif e.code() == grpc.StatusCode.UNAUTHENTICATED:
                    raise
                continue
            break
        return response

    return wrapper


class AggregatorGRPCClient:
    """Client to the aggregator over gRPC-TLS."""

    def __init__(
        self,
        agg_addr,
        agg_port,
        disable_client_auth,
        root_certificate,
        certificate,
        private_key,
        tls=True,
        aggregator_uuid=None,
        federation_uuid=None,
        single_col_cert_common_name=None,
        **kwargs,
    ):
        """Initialize."""
        self.uri = f"{agg_addr}:{agg_port}"
        self.tls = tls
        self.disable_client_auth = disable_client_auth
        self.root_certificate = root_certificate
        self.certificate = certificate
        self.private_key = private_key

        self.logger = getLogger(__name__)

        if not self.tls:
            self.logger.warn("gRPC is running on insecure channel with TLS disabled.")
            self.channel = self.create_insecure_channel(self.uri)
        else:
            self.channel = self.create_tls_channel(
                self.uri,
                self.root_certificate,
                self.disable_client_auth,
                self.certificate,
                self.private_key,
            )

        self.header = None
        self.aggregator_uuid = aggregator_uuid
        self.federation_uuid = federation_uuid
        self.single_col_cert_common_name = single_col_cert_common_name

        # Adding an interceptor for RPC Errors
        self.interceptors = (
            RetryOnRpcErrorClientInterceptor(
                sleeping_policy=ConstantBackoff(
                    logger=self.logger,
                    reconnect_interval=int(kwargs.get("client_reconnect_interval", 1)),
                    uri=self.uri,
                ),
                status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
            ),
        )
        self.stub = aggregator_pb2_grpc.AggregatorStub(
            grpc.intercept_channel(self.channel, *self.interceptors)
        )

    def create_insecure_channel(self, uri):
        """Set an insecure gRPC channel (i.e. no TLS) if desired.

        Warns user that this is not recommended.

        Args:
            uri: The uniform resource identifier of the insecure channel

        Returns:
            An insecure gRPC channel object
        """
        return grpc.insecure_channel(uri, options=channel_options)

    def create_tls_channel(
        self,
        uri,
        root_certificate,
        disable_client_auth,
        certificate,
        private_key,
    ):
        """Set an secure gRPC channel (i.e. TLS).

        Args:
            uri: The uniform resource identifier of the insecure channel
            root_certificate: The Certificate Authority filename
            disable_client_auth (boolean): True disabled client-side
             authentication (not recommended, throws warning to user)
            certificate: The client certificate filename from the collaborator
             (signed by the certificate authority)

        Returns:
            An insecure gRPC channel object
        """
        with open(root_certificate, "rb") as f:
            root_certificate_b = f.read()

        if disable_client_auth:
            self.logger.warn("Client-side authentication is disabled.")
            private_key_b = None
            certificate_b = None
        else:
            with open(private_key, "rb") as f:
                private_key_b = f.read()
            with open(certificate, "rb") as f:
                certificate_b = f.read()

        credentials = grpc.ssl_channel_credentials(
            root_certificates=root_certificate_b,
            private_key=private_key_b,
            certificate_chain=certificate_b,
        )

        return grpc.secure_channel(uri, credentials, options=channel_options)

    def _set_header(self, collaborator_name):
        self.header = aggregator_pb2.MessageHeader(
            sender=collaborator_name,
            receiver=self.aggregator_uuid,
            federation_uuid=self.federation_uuid,
            single_col_cert_common_name=self.single_col_cert_common_name or "",
        )

    def validate_response(self, reply, collaborator_name):
        """Validate the aggregator response."""
        assert reply.header.receiver == collaborator_name, (
            f"Receiver in response header does not match collaborator name. "
            f"Expected: {collaborator_name}, Actual: {reply.header.receiver}"
        )
        assert reply.header.sender == self.aggregator_uuid, (
            f"Sender in response header does not match aggregator UUID. "
            f"Expected: {self.aggregator_uuid}, Actual: {reply.header.sender}"
        )
        assert reply.header.federation_uuid == self.federation_uuid, (
            f"Federation UUID in response header does not match. "
            f"Expected: {self.federation_uuid}, Actual: {reply.header.federation_uuid}"
        )
        assert reply.header.single_col_cert_common_name == (
            self.single_col_cert_common_name or ""
        ), (
            f"Single collaborator certificate common name in response header does not match. "
            f"Expected: {self.single_col_cert_common_name or ''}, Actual: {reply.header.single_col_cert_common_name}"  # noqa: E501
        )

    def disconnect(self):
        """Close the gRPC channel."""
        self.logger.debug(f"Disconnecting from gRPC server at {self.uri}")
        self.channel.close()

    def reconnect(self):
        """Create a new channel with the gRPC server."""
        # channel.close() is idempotent. Call again here in case it wasn't
        # issued previously
        self.disconnect()

        if not self.tls:
            self.channel = self.create_insecure_channel(self.uri)
        else:
            self.channel = self.create_tls_channel(
                self.uri,
                self.root_certificate,
                self.disable_client_auth,
                self.certificate,
                self.private_key,
            )

        self.logger.debug(f"Connecting to gRPC at {self.uri}")

        self.stub = aggregator_pb2_grpc.AggregatorStub(
            grpc.intercept_channel(self.channel, *self.interceptors)
        )

    @_atomic_connection
    @_resend_data_on_reconnection
    def send_task_results(self, collaborator_name, round_number, next_step, clone_bytes):
        """Send next function name to aggregator."""
        self._set_header(collaborator_name)
        request = aggregator_pb2.TaskResultsRequest(
            header=self.header,
            collab_name=collaborator_name,
            round_number=round_number,
            next_step=next_step,
            execution_environment=clone_bytes,
        )

        response = self.stub.SendTaskResults(request)
        self.validate_response(response, collaborator_name)

        return response.header

    @_atomic_connection
    @_resend_data_on_reconnection
    def get_tasks(self, collaborator_name):
        """Get tasks from the aggregator."""
        self._set_header(collaborator_name)
        request = aggregator_pb2.GetTasksRequest(header=self.header)

        response = self.stub.GetTasks(request)
        self.validate_response(response, collaborator_name)

        return (
            response.round_number,
            response.function_name,
            response.execution_environment,
            response.sleep_time,
            response.quit,
        )

    @_atomic_connection
    @_resend_data_on_reconnection
    def call_checkpoint(self, collaborator_name, clone_bytes, function, stream_buffer):
        """Perform checkpoint for collaborator task."""
        self._set_header(collaborator_name)

        request = aggregator_pb2.CheckpointRequest(
            header=self.header,
            execution_environment=clone_bytes,
            function=function,
            stream_buffer=stream_buffer,
        )

        response = self.stub.CallCheckpoint(request)
        self.validate_response(response, collaborator_name)

        return response.header



# Copyright 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import ast
import inspect
import re
import sys
from importlib import import_module
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import nbformat
from nbdev.export import nb_export


class CodeAnalyzer:
    """Analyzes and process Jupyter Notebooks.
      Provides code extraction and transformation functionality

    Attributes:
        script_path (Path): Absolute path to the python script generated.
        script_name (str): Name of the generated python script.
        exported_script_module (ModuleType): The imported module object of the generated script.
        available_modules_in_exported_script (list): List of available attributes in the
            exported script.
    """

    def __init__(self, notebook_path: Path, output_path: Path) -> None:
        """Initialize CodeAnalyzer and process the script from notebook

        Args:
            notebook_path (Path): Path to Jupyter notebook to be converted.
            output_path (Path): The directory where the converted Python script will be saved.
        """
        print("Converting jupyter notebook to python script...")
        # Extract the export filename from the notebook
        self.script_name = self.__get_exp_name(notebook_path)
        # Convert the notebook to a Python script and set the script path
        self.script_path = Path(
            self.__convert_to_python(
                notebook_path,
                output_path.joinpath("src"),
                f"{self.script_name}.py",
            )
        ).resolve()
        self.__comment_flow_execution()

    def __get_exp_name(self, notebook_path: Path) -> str:
        """Extract experiment name from Jupyter notebook
        Looks for '#| default_exp <name>' pattern in code cells
        and extracts the experiment name. The name must be a valid Python identifier.

        Args:
            notebook_path (str): Path to Jupyter notebook.
        """
        with notebook_path.open("r") as f:
            notebook_content = nbformat.read(f, as_version=nbformat.NO_CONVERT)

        for cell in notebook_content.cells:
            if cell.cell_type == "code":
                code = cell.source
                match = re.search(r"#\s*\|\s*default_exp\s+(\w+)", code)
                if match:
                    print(f"Retrieved {match.group(1)} from default_exp")
                    return match.group(1)
        raise ValueError(
            "The notebook does not contain a '#| default_exp <experiment_name' marker."
            "Please add the marker to the first cell of the notebook"
        )

    def __convert_to_python(self, notebook_path: Path, output_path: Path, export_filename) -> Path:
        """Converts a Jupyter notebook to a Python script.
        Args:
            notebook_path (Path): The path to the Jupyter notebook file
                to be converted.
            output_path (Path): The directory where the exported Python
                script should be saved.
            export_filename: The name of the exported Python script file.

        Returns:
            Path: The path to the exported Python script file.
        """
        nb_export(notebook_path, output_path)

        return Path(output_path).joinpath(export_filename).resolve()

    def __comment_flow_execution(self) -> None:
        """Comment out lines containing '.run()' in the specified Python script"""
        run_statement = ".run()"

        with self.script_path.open("r") as f:
            data = f.readlines()
        for idx, line in enumerate(data):
            if run_statement in line:
                data[idx] = f"# {line}"
        with self.script_path.open("w") as f:
            f.writelines(data)

    def __import_generated_script(self) -> None:
        """
        Imports the generated python script using the importlib module
        """
        try:
            sys.path.append(str(self.script_path.parent))
            self.exported_script_module = import_module(self.script_name)
            self.available_modules_in_exported_script = dir(self.exported_script_module)
        except ImportError as e:
            raise ImportError(f"Failed to import script {self.script_name}: {e}")

    def __get_class_arguments(self, class_name) -> list:
        """Given the class name returns expected class arguments.

        Args:
            class_name (str): The name of the class.

        Returns:
            list: A list of expected class arguments.
        """
        if not hasattr(self, "exported_script_module"):
            self.__import_generated_script()

        # Find class from imported python script module
        for idx, attr in enumerate(self.available_modules_in_exported_script):
            if attr == class_name:
                cls = getattr(
                    self.exported_script_module,
                    self.available_modules_in_exported_script[idx],
                )
        if "cls" not in locals():
            raise NameError(f"{class_name} not found.")

        if inspect.isclass(cls):
            if "__init__" in cls.__dict__:
                init_signature = inspect.signature(cls.__init__)
                # Extract the parameter names (excluding 'self', 'args', and
                # 'kwargs')
                arg_names = [
                    param
                    for param in init_signature.parameters
                    if param not in ("self", "args", "kwargs")
                ]
                return arg_names
            return []
        print(f"{cls} is not a class")

    def __get_class_name(self, parent_class) -> Optional[str]:
        """Find and return the name of a class derived from the provided parent class.
        Args:
            parent_class: FLSpec instance.

        Returns:
            Optional[str]: The name of the derived class.
        """
        if not hasattr(self, "exported_script_module"):
            self.__import_generated_script()

        # Going though all attributes in imported python script
        for attr in self.available_modules_in_exported_script:
            t = getattr(self.exported_script_module, attr)
            if inspect.isclass(t) and t != parent_class and issubclass(t, parent_class):
                return attr
        raise ValueError("No flow class found that inherits from FLSpec")

    def __extract_class_initializing_args(self, class_name) -> Dict[str, Any]:
        """Provided name of the class returns expected arguments and it's
        values in form of dictionary.
        Args:
            class_name (str): The name of the class.

        Returns:
            Dict[str, Any]: A dictionary containing the expected arguments and their values.
        """
        instantiation_args = {"args": {}, "kwargs": {}}

        with open(self.script_path, "r") as s:
            tree = ast.parse(s.read())

            for node in ast.walk(tree):
                if isinstance(node, ast.Call) and isinstance(node.func, ast.Name):
                    if node.func.id == class_name:
                        # We found an instantiation of the class
                        instantiation_args["args"] = self._extract_positional_args(node.args)
                        instantiation_args["kwargs"] = self._extract_keyword_args(node.keywords)

        return instantiation_args

    def _extract_positional_args(self, args) -> Dict[str, Any]:
        """Extract positional arguments from the AST nodes.
        Args:
            args: AST nodes representing the arguments.

        Returns:
            Dict[str, Any]: Dictionary of argument names and their values.
        """
        positional_args = {}
        for arg in args:
            if isinstance(arg, ast.Name):
                positional_args[arg.id] = arg.id
            elif isinstance(arg, ast.Constant):
                positional_args[arg.s] = ast.unparse(arg)
            else:
                positional_args[arg.arg] = ast.unparse(arg).strip()
        return positional_args

    def _extract_keyword_args(self, keywords) -> Dict[str, Any]:
        """Extract keyword arguments from the AST nodes.
        Args:
            keywords: AST nodes representing the keyword arguments.

        Returns:
            Dict[str, Any]: Dictionary of keyword argument names and their values.
        """
        keyword_args = {}
        for kwarg in keywords:
            value = ast.unparse(kwarg.value).strip()
            value = self._clean_value(value)
            try:
                value = ast.literal_eval(value)
            except ValueError:
                pass
            keyword_args[kwarg.arg] = value
        return keyword_args

    def _clean_value(self, value: str) -> str:
        """Clean the value by removing unnecessary parentheses or brackets.
        Args:
            value (str): The string value to be cleaned.

        Returns:
            str: The cleaned string value
        """
        if value.startswith("(") and "," not in value:
            value = value.lstrip("(").rstrip(")")
        if value.startswith("[") and "," not in value:
            value = value.lstrip("[").rstrip("]")
        return value

    def get_requirements(self) -> Tuple[List[str], List[int], List[str]]:
        """Extract pip libraries from the script

        Returns:
            tuple: A tuple containing:
                requirements (list of str): List of pip libraries found in the script.
                line_nos (list of int): List of line numbers where "pip install" commands are found.
                data (list of str): The entire script data as a list of lines.
        """
        data = None
        with self.script_path.open("r") as f:
            requirements = []
            line_nos = []
            data = f.readlines()
            for i, line in enumerate(data):
                line = line.strip()
                if "pip install" in line:
                    line_nos.append(i)
                    # Avoid commented lines, libraries from *.txt file, or openfl.git
                    # installation
                    if not line.startswith("#") and "-r" not in line and "openfl.git" not in line:
                        requirements.append(f"{line.split(' ')[-1].strip()}\n")

            return requirements, line_nos, data

    def remove_lines(self, data: List[str], line_nos: List[int]) -> None:
        """Removes pip install lines from the script
        Args:
            data (List[str]): The entire script data as a list of lines.
            line_nos (List[int]): List of line numbers where "pip install" commands are found.
        """
        with self.script_path.open("w") as f:
            for i, line in enumerate(data):
                if i not in line_nos:
                    f.write(line)

    def get_flow_class_details(self, parent_class) -> Dict[str, Any]:
        """
        Retrieves details of a flow class that inherits from the given parent clas
        Args:
            parent_class: The parent class (FLSpec instance).

        Returns:
            Dict[str, Any]: A dictionary containing:
                flow_class_name (str): The name of the flow class.
                expected_args (List[str]): The expected arguments for the flow class.
                init_args (Dict[str, Any]): The initialization arguments for the flow class.
        """
        flow_class_name = self.__get_class_name(parent_class)
        expected_arguments = self.__get_class_arguments(flow_class_name)
        init_args = self.__extract_class_initializing_args(flow_class_name)

        return {
            "flow_class_name": flow_class_name,
            "expected_args": expected_arguments,
            "init_args": init_args,
        }

    def fetch_flow_configuration(self, flow_details: Dict[str, Any]) -> Dict[str, Any]:
        """Get flow configuration from flow details.
        Args:
            flow_details (Dict[str, Any]): Dictionary containing flow class details.

        Returns:
            Dict[str, Any]: Dictionary containing the plan configuration
        """
        flow_config = {
            "federated_flow": {
                "settings": {},
                "template": f"src.{self.script_name}.{flow_details['flow_class_name']}",
            }
        }

        def update_dictionary(args: dict, dtype: str = "args") -> None:
            """Update plan configuration with argument values.

            Args:
                args: Dictionary of arguments to process
                dtype: Type of arguments ('args' or 'kwargs')
            """
            for idx, (k, v) in enumerate(args.items()):
                if dtype == "args":
                    v = getattr(self.exported_script_module, str(k), None)
                    if v is not None and not isinstance(v, (int, str, bool)):
                        v = f"src.{self.script_name}.{k}"
                    k = flow_details["expected_args"][idx]
                elif dtype == "kwargs":
                    if v is not None and not isinstance(v, (int, str, bool)):
                        v = f"src.{self.script_name}.{v}"
                flow_config["federated_flow"]["settings"].update({k: v})

        # Process arguments
        pos_args = flow_details["init_args"].get("args", {})
        update_dictionary(pos_args, "args")
        kw_args = flow_details["init_args"].get("kwargs", {})
        update_dictionary(kw_args, "kwargs")

        return flow_config



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


class Callback:
    """Base class for callbacks.

    Callbacks can be used to perform actions at different stages of the
    Federated Learning process. To create a custom callback, subclass
    `openfl.callbacks.Callback` and implement the necessary methods.

    Callbacks can be triggered on the aggregator and collaborator side
    for the following events:
        * At the beginning of an experiment
        * At the beginning of a round
        * At the end of a round
        * At the end of an experiment

    Attributes:
        params: Additional parameters saved for use within the callback.
        tensor_db: The `TensorDB` instance of the respective participant.
    """

    def __init__(self):
        self.params = None
        self.tensor_db = None

    def set_params(self, params):
        self.params = params

    def set_tensor_db(self, tensor_db):
        self.tensor_db = tensor_db

    def on_round_begin(self, round_num: int, logs=None):
        """Callback function to be executed at the beginning of a round.

        Subclasses need to implement actions to be taken here.
        """

    def on_round_end(self, round_num: int, logs=None):
        """Callback function to be executed at the end of a round.

        Subclasses need to implement actions to be taken here.
        """

    def on_experiment_begin(self, logs=None):
        """Callback function to be executed at the beginning of an experiment.

        Subclasses need to implement actions to be taken here.
        """

    def on_experiment_end(self, logs=None):
        """Callback function to be executed at the end of an experiment.

        Subclasses need to implement actions to be taken here.
        """



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.utilities.data_splitters.data_splitter module."""

from abc import ABC, abstractmethod
from typing import Iterable, List, TypeVar

T = TypeVar("T")


class DataSplitter(ABC):
    """Base class for data splitting.

    This class should be subclassed when creating specific data splitter
    classes.
    """

    @abstractmethod
    def split(self, data: Iterable[T], num_collaborators: int) -> List[Iterable[T]]:
        """Split the data into a specified number of parts.

        Args:
            data (Iterable[T]): The data to be split.
            num_collaborators (int): The number of parts to split the data
                into.

        Returns:
            List[Iterable[T]]: The split data.

        Raises:
            NotImplementedError: This is an abstract method and must be
                overridden in a subclass.
        """
        raise NotImplementedError



# Copyright (C) 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import torch
import keras

class CNNModel(keras.Model):
    """
    A custom Keras model for a Convolutional Neural Network (CNN) that overrides
    the `train_step` and `test_step` methods to integrate with PyTorch's gradient
    computation and optimization.
    Methods
    -------
    train_step(data)
        Performs a single training step, including forward pass, loss computation,
        backward pass, and weight updates.
    test_step(data)
        Performs a single evaluation step, including forward pass, loss computation,
        and metric updates.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loss_tracker = keras.metrics.Mean(name="loss")
        self.mae_metric = keras.metrics.MeanAbsoluteError(name="mae")
        self.loss_fn = keras.losses.MeanSquaredError()

    def train_step(self, data):
        """
        Perform a single training step using torch.

        Args:
            data (tuple): A tuple containing the input data and labels. If the tuple has three elements,
                          the third element is considered as sample weights.

        Returns:
            dict: A dictionary mapping metric names to their current values, including the loss.

        The method performs the following steps:
        1. Unpacks the input data.
        2. Clears the gradients from the previous training step.
        3. Performs a forward pass to compute the predictions.
        4. Computes the loss based on the predictions and true labels.
        5. Computes the gradients by performing a backward pass on the loss.
        6. Updates the model weights using the computed gradients.
        7. Updates the metrics, including the loss.
        8. Returns the current values of the metrics.
        """

        # Unpack the data. Its structure depends on your model and
        # on what you pass to `fit()`.
        if len(data) == 3:
            x, y, sample_weight = data
        else:
            sample_weight = None
            x, y = data

        # Call torch.nn.Module.zero_grad() to clear the leftover gradients
        # for the weights from the previous train step.
        self.zero_grad()

        # Compute loss
        y_pred = self(x, training=True)  # Forward pass
        loss = self.compute_loss(
            y=y,
            y_pred=y_pred,
            sample_weight=sample_weight,
        )

        # Call torch.Tensor.backward() on the loss to compute gradients
        # for the weights.
        loss.backward()

        trainable_weights = [v for v in self.trainable_weights]
        gradients = [v.value.grad for v in trainable_weights]

        # Update weights
        with torch.no_grad():
            self.optimizer.apply(gradients, trainable_weights)

        # Update metrics (includes the metric that tracks the loss)
        for metric in self.metrics:
            if metric.name == "loss":
                metric.update_state(loss)
            else:
                metric.update_state(y, y_pred, sample_weight=sample_weight)

        # Return a dict mapping metric names to current value
        # Note that it will include the loss (tracked in self.metrics).
        return {m.name: m.result() for m in self.metrics}

    def test_step(self, data):
        """
        Perform a single test step using torch.
        Args:
            data (tuple): A tuple containing the input data (x) and the true labels (y).
        Returns:
            dict: A dictionary mapping metric names to their current values, including the loss.
        """

        # Unpack the data
        x, y = data
        # Compute predictions
        y_pred = self(x, training=False)
        # Updates the metrics tracking the loss
        loss = self.compute_loss(y=y, y_pred=y_pred)
        # Update the metrics.
        for metric in self.metrics:
            if metric.name == "loss":
                metric.update_state(loss)
            else:
                metric.update_state(y, y_pred)
        # Return a dict mapping metric names to current value.
        # Note that it will include the loss (tracked in self.metrics).
        return {m.name: m.result() for m in self.metrics}



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Experimental Experiment module."""

import asyncio
import logging
from contextlib import asynccontextmanager
from enum import Enum, auto
from pathlib import Path
from typing import Any, Iterable, List, Optional, Tuple, Union

from openfl.experimental.workflow.federated import Plan
from openfl.experimental.workflow.transport import AggregatorGRPCServer
from openfl.utilities.workspace import ExperimentWorkspace

logger = logging.getLogger(__name__)


class Status(Enum):
    """Experiment's statuses."""

    PENDING = auto()
    FINISHED = auto()
    IN_PROGRESS = auto()
    FAILED = auto()
    REJECTED = auto()


class Experiment:
    """Experiment class.

    Attributes:
            name (str): The name of the experiment.
            archive_path (Union[Path, str]): The path to the experiment
                archive.
            collaborators (List[str]): The list of collaborators.
            sender (str): The name of the sender.
            init_tensor_dict (dict): The initial tensor dictionary.
            plan_path (Union[Path, str]): The path to the plan.
            users (Iterable[str]): The list of users.
            status (str): The status of the experiment.
            aggregator (Aggregator): The aggregator instance.
            updated_flow (FLSpec): Updated flow instance.
    """

    def __init__(
        self,
        *,
        name: str,
        archive_path: Union[Path, str],
        collaborators: List[str],
        sender: str,
        plan_path: Union[Path, str] = "plan/plan.yaml",
        users: Iterable[str] = None,
    ) -> None:
        """Initialize an experiment object.

        Args:
            name (str): The name of the experiment.
            archive_path (Union[Path, str]): The path to the experiment
                archive.
            collaborators (List[str]): The list of collaborators.
            sender (str): The name of the sender.
            plan_path (Union[Path, str], optional): The path to the plan.
                Defaults to 'plan/plan.yaml'.
            users (Iterable[str], optional): The list of users. Defaults to
                None.
        """
        self.name = name
        self.archive_path = Path(archive_path).absolute()
        self.collaborators = collaborators
        self.sender = sender
        # This plan path ("plan/plan.yaml") originates from the
        # experiment workspace provided by the director
        self.plan_path = Path(plan_path)
        self.users = set() if users is None else set(users)
        self.status = Status.PENDING
        self.aggregator = None
        self.updated_flow = None

    async def start(
        self,
        *,
        tls: bool = True,
        root_certificate: Optional[Union[Path, str]] = None,
        private_key: Optional[Union[Path, str]] = None,
        certificate: Optional[Union[Path, str]] = None,
        director_config: Path = None,
        install_requirements: bool = False,
    ) -> Tuple[bool, Any]:
        """Run experiment.

        Args:
            tls (bool, optional): A flag indicating if TLS should be used for
                connections. Defaults to True.
            root_certificate (Optional[Union[Path, str]], optional): The path to the
                root certificate for TLS. Defaults to None.
            private_key (Optional[Union[Path, str]], optional): The path to the private
                key for TLS. Defaults to None.
            certificate (Optional[Union[Path, str]], optional): The path to the
                certificate for TLS. Defaults to None.
            director_config (Path): Path to director's config file
            install_requirements (bool, optional): A flag indicating if the
                requirements should be installed. Defaults to False.

        Returns:
            List[Union[bool, Any]]:
                - status: status of the experiment.
                - updated_flow: The updated flow object.
        """
        self.status = Status.IN_PROGRESS
        try:
            logger.info(f"New experiment {self.name} for collaborators {self.collaborators}")

            with ExperimentWorkspace(
                experiment_name=self.name,
                data_file_path=self.archive_path,
                install_requirements=install_requirements,
            ):
                aggregator_grpc_server = self._create_aggregator_grpc_server(
                    tls=tls,
                    root_certificate=root_certificate,
                    private_key=private_key,
                    certificate=certificate,
                    director_config=director_config,
                )
                self.aggregator = aggregator_grpc_server.aggregator
                _, self.updated_flow = await asyncio.gather(
                    self._run_aggregator_grpc_server(
                        aggregator_grpc_server,
                    ),
                    self.aggregator.run_flow(),
                )
            self.status = Status.FINISHED
            logger.info("Experiment %s was finished successfully.", self.name)
        except Exception as e:
            self.status = Status.FAILED
            logger.error("Experiment %s failed with error: %s.", self.name, e)
            raise

        return self.status == Status.FINISHED, self.updated_flow

    def _create_aggregator_grpc_server(
        self,
        *,
        tls: bool = True,
        root_certificate: Optional[Union[Path, str]] = None,
        private_key: Optional[Union[Path, str]] = None,
        certificate: Optional[Union[Path, str]] = None,
        director_config: Path = None,
    ) -> AggregatorGRPCServer:
        """Create an aggregator gRPC server.

        Args:
            tls (bool, optional): A flag indicating if TLS should be used for
                connections. Defaults to True.
            root_certificate (Optional[Union[Path, str]]): The path to the
                root certificate for TLS. Defaults to None.
            private_key (Optional[Union[Path, str]]): The path to the private
                key for TLS. Defaults to None.
            certificate (Optional[Union[Path, str]]): The path to the
                certificate for TLS. Defaults to None.
            director_config (Path): Path to director's config file.
                Defaults to None.
        Returns:
            AggregatorGRPCServer: The created aggregator gRPC server.
        """
        plan = Plan.parse(plan_config_path=self.plan_path)
        plan.authorized_cols = list(self.collaborators)

        logger.info("🧿 Created an Aggregator Server for %s experiment.", self.name)
        aggregator_grpc_server = plan.get_server(
            root_certificate=root_certificate,
            certificate=certificate,
            private_key=private_key,
            tls=tls,
            director_config=director_config,
        )
        return aggregator_grpc_server

    @staticmethod
    async def _run_aggregator_grpc_server(
        aggregator_grpc_server: AggregatorGRPCServer,
    ) -> None:
        """Run aggregator.

        Args:
            aggregator_grpc_server (AggregatorGRPCServer): The aggregator gRPC
                server to run.
        """
        logger.info("🧿 Starting the Aggregator Service.")
        grpc_server = aggregator_grpc_server.get_server()
        grpc_server.start()
        logger.info("Starting Aggregator gRPC Server")

        try:
            while not aggregator_grpc_server.aggregator.all_quit_jobs_sent():
                # Awaiting quit job sent to collaborators
                await asyncio.sleep(10)
            logger.debug("Aggregator sent quit jobs calls to all collaborators")
        except KeyboardInterrupt:
            logger.info("Keyboard interrupt received. Stopping the server.")
        finally:
            grpc_server.stop(0)


class ExperimentsRegistry:
    """ExperimentsList class."""

    def __init__(self) -> None:
        """Initialize an experiments registry object."""
        self.__active_experiment_name = None
        self.__pending_experiments = []
        self.__archived_experiments = []
        self.__dict = {}

    @property
    def active(self) -> Union[Experiment, None]:
        """Get active experiment.

        Returns:
            Union[Experiment, None]: The active experiment if exists, None
                otherwise.
        """
        if self.__active_experiment_name is None:
            return None
        return self.__dict[self.__active_experiment_name]

    @property
    def pending(self) -> List[str]:
        """Get queue of not started experiments.

        Returns:
            List[str]: The list of pending experiments.
        """
        return self.__pending_experiments

    def add(self, experiment: Experiment) -> None:
        """Add experiment to queue of not started experiments.

        Args:
           experiment (Experiment): The experiment to add.
        """
        self.__dict[experiment.name] = experiment
        self.__pending_experiments.append(experiment.name)

    def remove(self, name: str) -> None:
        """Remove experiment from everywhere.

        Args:
            name (str): The name of the experiment to remove.
        """
        if self.__active_experiment_name == name:
            self.__active_experiment_name = None
        if name in self.__pending_experiments:
            self.__pending_experiments.remove(name)
        if name in self.__archived_experiments:
            self.__archived_experiments.remove(name)
        if name in self.__dict:
            del self.__dict[name]

    def __getitem__(self, key: str) -> Experiment:
        """Get experiment by name.

        Args:
            key (str): The name of the experiment.

        Returns:
            Experiment: The experiment with the given name.
        """
        return self.__dict[key]

    def get(self, key: str, default=None) -> Experiment:
        """Get experiment by name.

        Args:
            key (str): The name of the experiment.
            default (optional): The default value to return if the experiment
                does not exist.

        Returns:
            Experiment: The experiment with the given name, or the default
                value if the experiment does not exist.
        """
        return self.__dict.get(key, default)

    def get_user_experiments(self, user: str) -> List[Experiment]:
        """Get list of experiments for specific user.

        Args:
            user (str): The name of the user.

        Returns:
            List[Experiment]: The list of experiments for the specific user.
        """
        return [exp for exp in self.__dict.values() if user in exp.users]

    def __contains__(self, key: str) -> bool:
        """Check if experiment exists.

        Args:
            key (str): The name of the experiment.

        Returns:
            bool: True if the experiment exists, False otherwise.
        """
        return key in self.__dict

    def finish_active(self) -> None:
        """Finish active experiment."""
        self.__archived_experiments.insert(0, self.__active_experiment_name)
        self.__active_experiment_name = None

    @asynccontextmanager
    async def get_next_experiment(self):
        """Context manager.

        On enter get experiment from pending experiments. On exit put finished
        experiment to archive_experiments.
        """
        while True:
            if self.active is None and self.pending:
                break
            await asyncio.sleep(10)

        try:
            self.__active_experiment_name = self.pending.pop(0)
            yield self.active
        finally:
            self.finish_active()



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
import os
import zipfile
from typing import Union

from openfl.federated import Plan
from openfl.federated.data.loader import DataLoader
from openfl.utilities.mocks import MockDataLoader


def get_dataloader(
    plan: Plan,
    prefer_minimal: bool = False,
    input_shape: Union[list, dict] = None,
    collaborator_index: int = 0,
) -> DataLoader:
    """Get dataloader instance from plan

    NOTE: if `prefer_minimal` is False, cwd must be the workspace directory
    because we need to construct dataloader from actual collaborator data path
    with actual data present.

    Args:
        plan (Plan):
            plan object linked with the dataloader
        prefer_minimal (bool ?):
            prefer to use MockDataLoader which can be used to more easily
            instantiate task_runner without any initial data.
            Default to `False`.
        input_shape (list | dict ?):
            input_shape denoted by list notation `[a,b,c, ...]` or in case
            of multihead models, dict object with individual layer keys such
            as `{"input_0": [a,b,...], "output_1": [x,y,z, ...]}`
            Defaults to `None`.
        collaborator_index (int ?):
            which collaborator should be used for initializing dataloader
            among collaborators specified in plan/data.yaml.
            Defaults to `0`.

    Returns:
        data_loader (DataLoader): DataLoader instance
    """

    # if specified, try to use minimal dataloader
    if prefer_minimal:
        # if input_shape not given, try to ascertain input_shape from plan
        if not input_shape and "input_shape" in plan.config["data_loader"]["settings"]:
            input_shape = plan.config["data_loader"]["settings"]["input_shape"]

        # input_shape is resolved; we can use the minimal dataloader intended
        # for util contexts which does not need a full dataloader with data
        if input_shape:
            data_loader: DataLoader = MockDataLoader(input_shape)
            # generically inherit all attributes from data_loader.settings
            for key, value in plan.config["data_loader"]["settings"].items():
                setattr(data_loader, key, value)
            return data_loader

    # Fallback; try to get a dataloader by constructing it from the collaborator
    # data directory path present in the the current workspace

    collaborator_names = list(plan.cols_data_paths)
    collatorators_count = len(collaborator_names)

    if collaborator_index >= collatorators_count:
        raise Exception(
            f"Unable to construct full dataloader from collab_index={collaborator_index} "
            f"when the plan has {collatorators_count} as total collaborator count. "
            f"Please check plan/data.yaml file for current collaborator entries."
        )

    collaborator_name = collaborator_names[collaborator_index]
    collaborator_data_path = plan.cols_data_paths[collaborator_name]

    # use seed_data provided by data_loader config if available
    if "seed_data" in plan.config["data_loader"]["settings"] and not os.path.isdir(
        collaborator_data_path
    ):
        os.makedirs(collaborator_data_path)
        sample_data_zip_file = plan.config["data_loader"]["settings"]["seed_data"]
        with zipfile.ZipFile(sample_data_zip_file, "r") as zip_ref:
            zip_ref.extractall(collaborator_data_path)

    data_loader = plan.get_data_loader(collaborator_name)

    return data_loader
